from itertools import chain

import numpy as np

from flaml.automl.data import pd
from flaml.automl.task.task import (
    MULTICHOICECLASSIFICATION,
    NLG_TASKS,
    SEQCLASSIFICATION,
    SEQREGRESSION,
    SUMMARIZATION,
    TOKENCLASSIFICATION,
)


def todf(X, Y, column_name):
    """
    todf converts Y from any format (list, pandas.Series, numpy array) to a DataFrame before being returned
    """
    if Y is not None:
        Y = pd.DataFrame(Y, index=X.index)
        Y.columns = column_name
    return Y


def tokenize_text(X, Y=None, task=None, hf_args=None, tokenizer=None):
    label_col_name = None
    # label_col_name is the name of the label column Y, label_col_name = ['labels'] for TOKENCLASSIFICATION and SUMMARIZATION,
    # label_col_name = ['label'] for other tasks. todf is used by all tasks except for SUMMARIZATION,
    # because the outputs of tokenize_seq2seq are already two DataFrames so no conversion needed.
    if task in (SEQCLASSIFICATION, SEQREGRESSION):
        X_tokenized = tokenize_onedataframe(
            X,
            tokenizer=tokenizer,
            task=task,
            hf_args=hf_args,
            prefix_str="",
        )
        Y_tokenized = Y
        label_col_name = ["label"]
    elif task == TOKENCLASSIFICATION:
        X_tokenized, Y_tokenized = tokenize_text_tokclassification(X, Y, tokenizer=tokenizer, hf_args=hf_args)
        label_col_name = ["labels"]
    elif task in NLG_TASKS:
        return tokenize_seq2seq(X, Y, tokenizer=tokenizer, task=task, hf_args=hf_args)
    elif task == MULTICHOICECLASSIFICATION:
        X_tokenized = tokenize_text_multiplechoice(X, tokenizer=tokenizer, hf_args=hf_args)
        label_col_name = ["label"]
        Y_tokenized = Y
    Y_tokenized = todf(X_tokenized, Y_tokenized, label_col_name)
    return X_tokenized, Y_tokenized


def tokenize_seq2seq(X, Y, tokenizer, task=None, hf_args=None):
    model_inputs = tokenize_onedataframe(
        X,
        tokenizer=tokenizer,
        task=task,
        hf_args=hf_args,
        prefix_str="summarize: ",
    )
    model_outputs = None
    if Y is not None:
        model_outputs = tokenize_onedataframe(
            Y.to_frame(),
            tokenizer=tokenizer,
            task=task,
            hf_args=hf_args,
            prefix_str="",
        )
        model_outputs["labels"] = [
            [(each_l if each_l != tokenizer.pad_token_id else -100) for each_l in label]
            for label in model_outputs["input_ids"]
        ]
        model_outputs = model_outputs.drop(columns=["attention_mask", "input_ids", "decoder_input_ids"])
    return model_inputs, model_outputs


def tokenize_and_align_labels(
    examples,
    tokenizer,
    label_to_id,
    b_to_i_label,
    hf_args=None,
    X_sent_key=None,
    Y_sent_key=None,
    return_column_name=False,
):
    # tokenize_and_align_labels is only called by the token-classification task
    tokenized_inputs = tokenizer(
        [list(examples[X_sent_key])],
        padding="max_length"
        if hf_args and hf_args.pad_to_max_length
        else False,  # to be consistent with https://github.com/huggingface/transformers/blob/main/examples/pytorch/token-classification/run_ner.py#L394
        truncation=True,
        max_length=hf_args.max_seq_length if hf_args else None,
        # We use this argument because the texts in our dataset are lists of words (with a label for each word).
        is_split_into_words=True,
    )
    if Y_sent_key is not None:
        previous_word_idx = None
        label_ids = []
        for word_idx in tokenized_inputs.word_ids(batch_index=0):
            if word_idx is None:
                label_ids.append(-100)
            elif word_idx != previous_word_idx:
                label_ids.append(label_to_id[examples[Y_sent_key][word_idx]])
            # For the other tokens in a word, we set the label to either the current label or -100, depending on
            # the label_all_tokens flag.
            else:
                # Use the label_all_tokens to control whether to copy the label to all subtokens or to pad the additional tokens as -100
                if hf_args.label_all_tokens:
                    # If the B- word is converted into multiple subtokens, map the additional subtokens to I-
                    label_ids.append(b_to_i_label[label_to_id[examples[Y_sent_key][word_idx]]])
                else:
                    label_ids.append(-100)
            previous_word_idx = word_idx
        tokenized_inputs["labels"] = label_ids
    tmp_column_names = sorted(tokenized_inputs.keys())
    tokenized_input_and_labels = [tokenized_inputs[x] for x in tmp_column_names]
    for key_idx, each_key in enumerate(tmp_column_names):
        if each_key != "labels":
            tokenized_input_and_labels[key_idx] = tokenized_input_and_labels[key_idx][0]
    if return_column_name:
        return tokenized_input_and_labels, tmp_column_names
    else:
        return tokenized_input_and_labels


def tokenize_text_tokclassification(X, Y, tokenizer, hf_args=None):
    # If the label_all_tokens flag is True, prepare two dicts label_to_id and b_to_i_label to convert the B- labels to I- labels
    label_to_id = {i: i for i in range(len(hf_args.label_list))}
    b_to_i_label = []
    for idx, label in enumerate(hf_args.label_list):
        if label.startswith("B-") and label.replace("B-", "I-") in hf_args.label_list:
            b_to_i_label.append(hf_args.label_list.index(label.replace("B-", "I-")))
        else:
            b_to_i_label.append(idx)

    if Y is not None:
        X_and_Y = pd.concat([X, Y.to_frame()], axis=1)
        X_key = list(X.keys())[0]
        Y_key = list(Y.to_frame().keys())[0]
        # tokenize_and_align_labels is only called by the token-classification task
        _, tokenized_column_names = tokenize_and_align_labels(
            X_and_Y.iloc[0],
            tokenizer=tokenizer,
            hf_args=hf_args,
            X_sent_key=X_key,
            Y_sent_key=Y_key,
            return_column_name=True,
            label_to_id=label_to_id,
            b_to_i_label=b_to_i_label,
        )
        X_and_Y_tokenized = X_and_Y.apply(
            lambda x: tokenize_and_align_labels(
                x,
                tokenizer=tokenizer,
                hf_args=hf_args,
                X_sent_key=X_key,
                Y_sent_key=Y_key,
                label_to_id=label_to_id,
                b_to_i_label=b_to_i_label,
            ),
            axis=1,
            result_type="expand",
        )
        label_idx = tokenized_column_names.index("labels")
        other_indices = sorted(set(range(len(tokenized_column_names))).difference({label_idx}))
        other_column_names = [tokenized_column_names[x] for x in other_indices]
        d = X_and_Y_tokenized.iloc[:, other_indices]
        y_tokenized = X_and_Y_tokenized.iloc[:, label_idx]
    else:
        X_key = list(X.keys())[0]

        _, tokenized_column_names = tokenize_and_align_labels(
            X.iloc[0],
            tokenizer=tokenizer,
            hf_args=hf_args,
            X_sent_key=X_key,
            Y_sent_key=None,
            return_column_name=True,
            label_to_id=label_to_id,
            b_to_i_label=b_to_i_label,
        )

        d = X.apply(
            lambda x: tokenize_and_align_labels(
                x,
                tokenizer=tokenizer,
                hf_args=hf_args,
                X_sent_key=X_key,
                Y_sent_key=None,
                label_to_id=label_to_id,
                b_to_i_label=b_to_i_label,
            ),
            axis=1,
            result_type="expand",
        )
        other_column_names = tokenized_column_names
        y_tokenized = None
    X_tokenized = pd.DataFrame(columns=other_column_names)
    X_tokenized[other_column_names] = d
    return X_tokenized, y_tokenized


def tokenize_onedataframe(
    X,
    tokenizer,
    task=None,
    hf_args=None,
    prefix_str=None,
):
    with tokenizer.as_target_tokenizer():
        _, tokenized_column_names = tokenize_row(
            dict(X.iloc[0]),
            tokenizer,
            prefix=(prefix_str,) if task is SUMMARIZATION else None,
            task=task,
            hf_args=hf_args,
            return_column_name=True,
        )
        d = X.apply(
            lambda x: tokenize_row(
                x,
                tokenizer,
                prefix=(prefix_str,) if task is SUMMARIZATION else None,
                task=task,
                hf_args=hf_args,
            ),
            axis=1,
            result_type="expand",
        )
        X_tokenized = pd.DataFrame(columns=tokenized_column_names)
        X_tokenized[tokenized_column_names] = d
        return X_tokenized


def tokenize_row(
    this_row,
    tokenizer,
    prefix=None,
    task=None,
    hf_args=None,
    return_column_name=False,
):
    if prefix:
        this_row = tuple("".join(x) for x in zip(prefix, this_row))

    # tokenizer.pad_token = tokenizer.eos_token
    tokenized_example = tokenizer(
        *tuple(this_row),
        padding="max_length" if hf_args and hf_args.pad_to_max_length else False,
        max_length=hf_args.max_seq_length if hf_args else None,
        truncation=True,
    )
    if task in NLG_TASKS:
        tokenized_example["decoder_input_ids"] = tokenized_example["input_ids"]
    tmp_column_names = sorted(tokenized_example.keys())

    if return_column_name:
        return [tokenized_example[x] for x in tmp_column_names], tmp_column_names
    else:
        return [tokenized_example[x] for x in tmp_column_names]


def tokenize_text_multiplechoice(X, tokenizer, hf_args=None):
    t = X[["sent1", "sent2", "ending0", "ending1", "ending2", "ending3"]]
    _, tokenized_column_names = tokenize_swag(
        t.iloc[0],
        tokenizer=tokenizer,
        hf_args=hf_args,
        return_column_name=True,
    )
    d = t.apply(
        lambda x: tokenize_swag(x, tokenizer=tokenizer, hf_args=hf_args),
        axis=1,
        result_type="expand",
    )

    X_tokenized = pd.DataFrame(columns=tokenized_column_names)
    X_tokenized[tokenized_column_names] = d
    output = X_tokenized.join(X)
    return output


def tokenize_swag(this_row, tokenizer, hf_args=None, return_column_name=False):
    first_sentences = [[this_row["sent1"]] * 4]
    # get each 1st sentence, multiply to 4 sentences
    question_headers = this_row["sent2"]
    # sent2 are the noun part of 2nd line
    second_sentences = [question_headers + " " + this_row[key] for key in ["ending0", "ending1", "ending2", "ending3"]]
    # now the 2nd-sentences are formed by combing the noun part and 4 ending parts

    # Flatten out
    # From 2 dimension to 1 dimension array
    first_sentences = list(chain(*first_sentences))

    tokenized_example = tokenizer(
        *tuple([first_sentences, second_sentences]),
        truncation=True,
        max_length=hf_args.max_seq_length if hf_args else None,
        padding="max_length" if hf_args and hf_args.pad_to_max_length else False,
    )
    tmp_column_names = sorted(tokenized_example.keys())

    if return_column_name:
        return [tokenized_example[x] for x in tmp_column_names], tmp_column_names
    else:
        return [tokenized_example[x] for x in tmp_column_names]


def postprocess_prediction_and_true(task, y_pred, tokenizer, hf_args, y_true=None, X=None):
    # postprocess the matrix prediction y_pred and ground truth y_true into user readable format, e.g., for summarization, decode into text
    if y_pred is None:
        return np.array([0.0] * len(X)), y_true
    if task == SEQCLASSIFICATION:
        return np.argmax(y_pred, axis=1), y_true
    elif task == SEQREGRESSION:
        return np.squeeze(y_pred), y_true  # predictions.reshape((len(predictions),))
    elif task == TOKENCLASSIFICATION:
        assert (y_true is not None) or (X is not None), "One of y_true and X must not be None"
        ## If y_true is not None, we use y_true to remove the -100 in the prediction (postprocessing), and return the postprocessed y_true and prediction
        # If y_true is None, we use X to compute y_is_pad (i.e., whether y_true is -100 in that position), and use y_is_pad to remove the -100 in the prediction, and return the postprocessed prediction (not the y_true)
        y_predict = pd.Series(np.argmax(y_pred, axis=2).tolist())
        if y_true is None:
            _, y_is_pad_df = tokenize_text(
                X,
                y_predict,
                task=task,
                hf_args=hf_args,
                tokenizer=tokenizer,
            )
            y_is_pad = y_is_pad_df.iloc[:, 0]
        else:
            y_is_pad = y_true
        label_len = len(hf_args.label_list)
        zip_pred_ispad = [
            [(p, ispd) for (p, ispd) in zip(each_pred, each_is_pad) if ispd != -100]
            for (each_pred, each_is_pad) in zip(y_predict, y_is_pad)
        ]
        y_pred_label = [
            [hf_args.label_list[p] if 0 <= p < label_len else -1 for (p, ispd) in each_list]
            for each_list in zip_pred_ispad
        ]  # To compute precision and recall, y_pred and y_true must be converted to string labels
        # (B-PER, I-PER, etc.), so that the category-based precision/recall (i.e., PER, LOC, etc.) scores can be computed
        if y_true is not None:
            y_true_label = [[tr for (p, tr) in each_list] for each_list in zip_pred_ispad]
        else:
            y_true_label = None
        return y_pred_label, y_true_label
    elif task == SUMMARIZATION:
        if isinstance(y_pred, tuple):
            y_pred = np.argmax(y_pred[0], axis=2)
        decoded_preds = tokenizer.batch_decode(y_pred, skip_special_tokens=True)

        import nltk

        nltk.download("punkt")
        decoded_preds = [pred.strip() for pred in decoded_preds]
        decoded_preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in decoded_preds]

        if y_true is not None:
            y_true_labels = np.where(y_true != -100, y_true, tokenizer.pad_token_id)
            decoded_y_true_labels = tokenizer.batch_decode(y_true_labels, skip_special_tokens=True)
            decoded_y_true_labels = [label.strip() for label in decoded_y_true_labels]
            decoded_y_true_labels = ["\n".join(nltk.sent_tokenize(label)) for label in decoded_y_true_labels]
        else:
            decoded_y_true_labels = None

        return decoded_preds, decoded_y_true_labels
    elif task == MULTICHOICECLASSIFICATION:
        return np.argmax(y_pred, axis=1), y_true


def load_model(checkpoint_path, task, num_labels=None):
    import transformers

    transformers.logging.set_verbosity_error()

    from transformers import AutoConfig

    from flaml.automl.task.task import (
        SEQCLASSIFICATION,
        SEQREGRESSION,
        TOKENCLASSIFICATION,
    )

    def get_this_model(checkpoint_path, task, model_config):
        from transformers import (
            AutoModelForMultipleChoice,
            AutoModelForSeq2SeqLM,
            AutoModelForSequenceClassification,
            AutoModelForTokenClassification,
        )

        if task in (SEQCLASSIFICATION, SEQREGRESSION):
            return AutoModelForSequenceClassification.from_pretrained(
                checkpoint_path, config=model_config, ignore_mismatched_sizes=True
            )
        elif task == TOKENCLASSIFICATION:
            return AutoModelForTokenClassification.from_pretrained(checkpoint_path, config=model_config)
        elif task in NLG_TASKS:
            return AutoModelForSeq2SeqLM.from_pretrained(checkpoint_path, config=model_config)
        elif task == MULTICHOICECLASSIFICATION:
            return AutoModelForMultipleChoice.from_pretrained(checkpoint_path, config=model_config)

    def _set_model_config(checkpoint_path):
        if task in (SEQCLASSIFICATION, SEQREGRESSION, TOKENCLASSIFICATION):
            model_config = AutoConfig.from_pretrained(
                checkpoint_path,
                num_labels=model_config_num_labels,
            )
            return model_config
        else:
            model_config = AutoConfig.from_pretrained(checkpoint_path)
            return model_config

    current_config = AutoConfig.from_pretrained(checkpoint_path)
    this_vocab_size = current_config.vocab_size

    model_config_num_labels = num_labels
    new_config = _set_model_config(checkpoint_path)

    this_model = get_this_model(checkpoint_path, task, new_config)
    this_model.resize_token_embeddings(this_vocab_size)
    return this_model



try:
    from sklearn.ensemble import HistGradientBoostingClassifier, HistGradientBoostingRegressor
except ImportError as e:
    print(f"scikit-learn is required for HistGradientBoostingEstimator. Please install it; error: {e}")

from flaml import tune
from flaml.automl.model import SKLearnEstimator
from flaml.automl.task import Task


class HistGradientBoostingEstimator(SKLearnEstimator):
    """The class for tuning Histogram Gradient Boosting."""

    ITER_HP = "max_iter"
    HAS_CALLBACK = False
    DEFAULT_ITER = 100

    @classmethod
    def search_space(cls, data_size: int, task, **params) -> dict:
        upper = max(5, min(32768, int(data_size[0])))  # upper must be larger than lower
        return {
            "n_estimators": {
                "domain": tune.lograndint(lower=4, upper=upper),
                "init_value": 4,
                "low_cost_init_value": 4,
            },
            "max_leaves": {
                "domain": tune.lograndint(lower=4, upper=upper),
                "init_value": 4,
                "low_cost_init_value": 4,
            },
            "min_samples_leaf": {
                "domain": tune.lograndint(lower=2, upper=2**7 + 1),
                "init_value": 20,
            },
            "learning_rate": {
                "domain": tune.loguniform(lower=1 / 1024, upper=1.0),
                "init_value": 0.1,
            },
            "log_max_bin": {  # log transformed with base 2, <= 256
                "domain": tune.lograndint(lower=3, upper=9),
                "init_value": 8,
            },
            "l2_regularization": {
                "domain": tune.loguniform(lower=1 / 1024, upper=1024),
                "init_value": 1.0,
            },
        }

    def config2params(self, config: dict) -> dict:
        params = super().config2params(config)
        if "log_max_bin" in params:
            params["max_bins"] = (1 << params.pop("log_max_bin")) - 1
        if "max_leaves" in params:
            params["max_leaf_nodes"] = params.get("max_leaf_nodes", params.pop("max_leaves"))
        if "n_estimators" in params:
            params["max_iter"] = params.get("max_iter", params.pop("n_estimators"))
        if "random_state" not in params:
            params["random_state"] = 24092023
        if "n_jobs" in params:
            params.pop("n_jobs")
        return params

    def __init__(
        self,
        task: Task,
        **config,
    ):
        super().__init__(task, **config)
        self.params["verbose"] = 0

        if self._task.is_classification():
            self.estimator_class = HistGradientBoostingClassifier
        else:
            self.estimator_class = HistGradientBoostingRegressor



import os

os.environ["PYARROW_IGNORE_TIMEZONE"] = "1"
try:
    import pyspark
    import pyspark.pandas as ps
    import pyspark.sql.functions as F
    import pyspark.sql.types as T
    from pyspark.pandas import DataFrame as psDataFrame
    from pyspark.pandas import Series as psSeries
    from pyspark.pandas import set_option
    from pyspark.sql import DataFrame as sparkDataFrame
    from pyspark.util import VersionUtils
except ImportError:

    class psDataFrame:
        pass

    F = T = ps = sparkDataFrame = psSeries = psDataFrame
    _spark_major_minor_version = set_option = None
    ERROR = ImportError(
        """Please run pip install flaml[spark]
    and check [here](https://spark.apache.org/docs/latest/api/python/getting_started/install.html)
    for more details about installing Spark."""
    )
else:
    ERROR = None
    _spark_major_minor_version = VersionUtils.majorMinorVersion(pyspark.__version__)

try:
    import pandas as pd
    from pandas import DataFrame, Series
except ImportError:
    DataFrame = Series = pd = None



import atexit
import functools
import json
import logging
import os
import pickle
import random
import tempfile
import time
import warnings
from concurrent.futures import ThreadPoolExecutor, wait
from typing import MutableMapping

import mlflow
import pandas as pd
from mlflow.entities import Metric, Param, RunTag
from mlflow.exceptions import MlflowException
from mlflow.utils.autologging_utils import AUTOLOGGING_INTEGRATIONS, autologging_is_disabled
from packaging.requirements import Requirement
from scipy.sparse import issparse
from sklearn import tree

try:
    from pyspark.ml import PipelineModel as SparkPipelineModel
except ImportError:

    class SparkPipelineModel:
        pass


# from mlflow.store.tracking import SEARCH_MAX_RESULTS_THRESHOLD
from sklearn.pipeline import Pipeline

from flaml.automl.logger import logger
from flaml.automl.spark import DataFrame, Series, psDataFrame, psSeries
from flaml.version import __version__

SEARCH_MAX_RESULTS = 5000  # Each train should not have more than 5000 trials
IS_RENAME_CHILD_RUN = os.environ.get("FLAML_IS_RENAME_CHILD_RUN", "false").lower() == "true"
REMOVE_REQUIREMENT_LIST = [
    "synapseml-cognitive",
    "synapseml-core",
    "synapseml-deep-learning",
    "synapseml-internal",
    "synapseml-mlflow",
    "synapseml-opencv",
    "synapseml-vw",
    "synapseml-lightgbm",
    "synapseml-utils",
    "nni",
    "optuna",
]
OPTIONAL_REMOVE_REQUIREMENT_LIST = ["pytorch-lightning", "transformers"]

os.environ["MLFLOW_ENABLE_ARTIFACTS_PROGRESS_BAR"] = os.environ.get("MLFLOW_ENABLE_ARTIFACTS_PROGRESS_BAR", "false")

MLFLOW_NUM_WORKERS = int(os.environ.get("FLAML_MLFLOW_NUM_WORKERS", os.cpu_count() * 4 if os.cpu_count() else 2))
executor = ThreadPoolExecutor(max_workers=MLFLOW_NUM_WORKERS)
atexit.register(lambda: executor.shutdown(wait=True))

IS_CLEAN_LOGS = os.environ.get("FLAML_IS_CLEAN_LOGS", "1")
if IS_CLEAN_LOGS == "1":
    logging.getLogger("synapse.ml").setLevel(logging.CRITICAL)
    logging.getLogger("mlflow.utils").setLevel(logging.CRITICAL)
    logging.getLogger("mlflow.utils.environment").setLevel(logging.CRITICAL)
    logging.getLogger("mlflow.models.model").setLevel(logging.CRITICAL)
    warnings.simplefilter("ignore", category=FutureWarning)
    warnings.simplefilter("ignore", category=UserWarning)


def convert_requirement(requirement_list: list[str]):
    ret = (
        [Requirement(s.strip().lower()) for s in requirement_list]
        if mlflow.__version__ <= "2.17.0"
        else requirement_list
    )
    return ret


def time_it(func_or_code=None):
    """
    Decorator or function that measures execution time.

    Can be used in three ways:
    1. As a decorator with no arguments: @time_it
    2. As a decorator with arguments: @time_it()
    3. As a function call with a string of code to execute and time: time_it("some_code()")

    Args:
        func_or_code (callable or str, optional): Either a function to decorate or
            a string of code to execute and time.

    Returns:
        callable or None: Returns a decorated function if used as a decorator,
            or None if used to execute a string of code.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result = func(*args, **kwargs)
            end_time = time.time()
            logger.debug(f"Execution of {func.__name__} took {end_time - start_time:.4f} seconds")
            return result

        return wrapper

    if callable(func_or_code):
        return decorator(func_or_code)
    elif func_or_code is None:
        return decorator
    else:
        start_time = time.time()
        exec(func_or_code)
        end_time = time.time()
        logger.debug(f"Execution\n```\n{func_or_code}\n```\ntook {end_time - start_time:.4f} seconds")


def flatten_dict(d: MutableMapping, sep: str = ".") -> MutableMapping:
    if len(d) == 0:
        return d
    [flat_dict] = pd.json_normalize(d, sep=sep).to_dict(orient="records")
    keys = list(flat_dict.keys())
    for key in keys:
        if not isinstance(flat_dict[key], (int, float)):
            flat_dict.pop(key)
    return flat_dict


def is_autolog_enabled():
    return not all(autologging_is_disabled(k) for k in AUTOLOGGING_INTEGRATIONS.keys())


def get_mlflow_log_latency(model_history=False, delete_run=True):
    try:
        FLAML_MLFLOW_LOG_LATENCY = float(os.getenv("FLAML_MLFLOW_LOG_LATENCY", 0))
    except ValueError:
        FLAML_MLFLOW_LOG_LATENCY = 0
    if FLAML_MLFLOW_LOG_LATENCY >= 0.1:
        return FLAML_MLFLOW_LOG_LATENCY
    st = time.time()
    with mlflow.start_run(nested=True, run_name="get_mlflow_log_latency") as run:
        if model_history:
            sk_model = tree.DecisionTreeClassifier()
            mlflow.sklearn.log_model(sk_model, "model")
            with tempfile.TemporaryDirectory() as tmpdir:
                pickle_fpath = os.path.join(tmpdir, f"tmp_{int(time.time() * 1000)}")
                with open(pickle_fpath, "wb") as f:
                    pickle.dump(sk_model, f)
                mlflow.log_artifact(pickle_fpath, "sk_model")
        mlflow.set_tag("synapseml.ui.visible", "false")  # not shown inline in fabric
    if delete_run:
        mlflow.delete_run(run.info.run_id)
    et = time.time()
    return 3 * (et - st)


def infer_signature(X_train=None, y_train=None, dataframe=None, label=None):
    if X_train is not None:
        if issparse(X_train):
            X_train = X_train.tocsr()
        elif isinstance(X_train, psDataFrame):
            X_train = X_train.to_spark(index_col="tmp_index_col")
            y_train = None
        try:
            signature = mlflow.models.infer_signature(X_train, y_train)
            return signature
        except (TypeError, MlflowException, Exception) as e:
            logger.debug(
                f"Failed to infer signature from X_train {type(X_train)} and y_train {type(y_train)}, error: {e}"
            )
    else:
        if dataframe is not None and label is not None:
            X = dataframe.drop(columns=label)
            y = dataframe[label]
            if isinstance(dataframe, psDataFrame):
                X = X.to_spark(index_col="tmp_index_col")
                y = None
            try:
                signature = mlflow.models.infer_signature(X, y)
                return signature
            except (TypeError, MlflowException, Exception) as e:
                logger.debug(
                    f"Failed to infer signature from dataframe {type(dataframe)} and label {label}, error: {e}"
                )


def update_and_install_requirements(
    run_id=None,
    model_name=None,
    model_version=None,
    remove_list=None,
    artifact_path="model",
    dst_path=None,
    install_with_ipython=False,
):
    if not (run_id or (model_name and model_version)):
        raise ValueError(
            "Please provide `run_id` or both `model_name` and `model_version`. If all three are provided, `run_id` will be used."
        )

    if install_with_ipython:
        from IPython import get_ipython

    if not remove_list:
        remove_list = [
            "synapseml-cognitive",
            "synapseml-core",
            "synapseml-deep-learning",
            "synapseml-internal",
            "synapseml-mlflow",
            "synapseml-opencv",
            "synapseml-vw",
            "synapseml-lightgbm",
            "synapseml-utils",
            "flaml",  # flaml is needed for AutoML models, should be pre-installed in the runtime
            "pyspark",  # fabric internal pyspark should be pre-installed in the runtime
        ]

    # Download model artifacts
    client = mlflow.MlflowClient()
    if not run_id:
        run_id = client.get_model_version(model_name, model_version).run_id
    if not dst_path:
        dst_path = os.path.join(tempfile.gettempdir(), "model_artifacts")
    os.makedirs(dst_path, exist_ok=True)
    client.download_artifacts(run_id, artifact_path, dst_path)
    requirements_path = os.path.join(dst_path, artifact_path, "requirements.txt")
    with open(requirements_path) as f:
        reqs = f.read().splitlines()
        old_reqs = [Requirement(req) for req in reqs if req]
        old_reqs_dict = {req.name: str(req) for req in old_reqs}
        for req in remove_list:
            req = Requirement(req)
            if req.name in old_reqs_dict:
                old_reqs_dict.pop(req.name, None)
        new_reqs_list = list(old_reqs_dict.values())

    with open(requirements_path, "w") as f:
        f.write("\n".join(new_reqs_list))

    if install_with_ipython:
        get_ipython().run_line_magic("pip", f"install -r {requirements_path} -q")
    else:
        logger.info(f"You can run `pip install -r {requirements_path}` to install dependencies.")
    return requirements_path


def _mlflow_wrapper(evaluation_func, mlflow_exp_id, mlflow_config=None, extra_tags=None, autolog=False):
    def wrapped(*args, **kwargs):
        if mlflow_config is not None:
            try:
                from synapse.ml.mlflow import set_mlflow_env_config

                set_mlflow_env_config(mlflow_config)
            except Exception:
                pass
        import mlflow

        if mlflow_exp_id is not None:
            mlflow.set_experiment(experiment_id=mlflow_exp_id)
        if autolog:
            if mlflow.__version__ > "2.5.0" and extra_tags is not None:
                mlflow.autolog(silent=True, extra_tags=extra_tags)
            else:
                mlflow.autolog(silent=True)
            logger.debug("activated mlflow autologging on executor")
        else:
            mlflow.autolog(disable=True, silent=True)
        # with mlflow.start_run(nested=True):
        result = evaluation_func(*args, **kwargs)
        return result

    return wrapped


def _get_notebook_name():
    try:
        import re

        from synapse.ml.mlflow import get_mlflow_env_config
        from synapse.ml.mlflow.shared_platform_utils import get_artifact

        notebook_id = get_mlflow_env_config(False).artifact_id
        current_notebook = get_artifact(notebook_id)
        notebook_name = re.sub("\\W+", "-", current_notebook.displayName).strip()

        return notebook_name
    except Exception as e:
        logger.debug(f"Failed to get notebook name: {e}")
        return None


def safe_json_dumps(obj):
    def default(o):
        return str(o)

    return json.dumps(obj, default=default)


class MLflowIntegration:
    def __init__(self, experiment_type="automl", mlflow_exp_name=None, extra_tag=None):
        try:
            from synapse.ml.mlflow import get_mlflow_env_config

            self.driver_mlflow_env_config = get_mlflow_env_config()
            self._on_internal = True
            self._notebook_name = _get_notebook_name()
        except ModuleNotFoundError:
            self.driver_mlflow_env_config = None
            self._on_internal = False
            self._notebook_name = None

        self.autolog = False
        self.manual_log = False
        self.parent_run_id = None
        self.parent_run_name = None
        self.log_type = "null"
        self.resume_params = {}
        self.train_func = None
        self.best_iteration = None
        self.best_run_id = None
        self.child_counter = 0
        self.infos = []
        self.manual_run_ids = []
        self.has_summary = False
        self.has_model = False
        self.only_history = False
        self._do_log_model = True
        self.futures = {}
        self.futures_log_model = {}

        self.extra_tag = (
            extra_tag
            if extra_tag is not None
            else {"extra_tag.sid": f"flaml_{__version__}_{int(time.time())}_{random.randint(1001, 9999)}"}
        )
        self.start_time = time.time()
        self.experiment_type = experiment_type
        self.update_autolog_state()

        self.mlflow_client = mlflow.tracking.MlflowClient()
        parent_run_info = mlflow.active_run().info if mlflow.active_run() is not None else None
        if parent_run_info:
            self.experiment_id = parent_run_info.experiment_id
            self.parent_run_id = parent_run_info.run_id
            # attribute run_name is not available before mlflow 2.0.1
            self.parent_run_name = parent_run_info.run_name if hasattr(parent_run_info, "run_name") else "flaml_run"
            if self.parent_run_name == "":
                self.parent_run_name = mlflow.active_run().data.tags["mlflow.runName"]
        else:
            if mlflow_exp_name is None:
                if mlflow.tracking.fluent._active_experiment_id is None:
                    mlflow_exp_name = self._notebook_name if self._notebook_name else "flaml_default_experiment"
                    mlflow.set_experiment(experiment_name=mlflow_exp_name)
            else:
                mlflow.set_experiment(experiment_name=mlflow_exp_name)
            self.experiment_id = mlflow.tracking.fluent._active_experiment_id
        self.experiment_name = mlflow.get_experiment(self.experiment_id).name

        if self.autolog:
            # only end user created parent run in autolog scenario
            mlflow.end_run()

    def set_mlflow_config(self):
        if self.driver_mlflow_env_config is not None:
            try:
                from synapse.ml.mlflow import set_mlflow_env_config

                set_mlflow_env_config(self.driver_mlflow_env_config)
            except Exception:
                pass

    def wrap_evaluation_function(self, evaluation_function):
        wrapped_evaluation_function = _mlflow_wrapper(
            evaluation_function, self.experiment_id, self.driver_mlflow_env_config, self.extra_tag, self.autolog
        )
        return wrapped_evaluation_function

    def set_best_iter(self, result):
        # result: AutoML or ExperimentAnalysis
        try:
            self.best_iteration = result.best_iteration
        except AttributeError:
            self.best_iteration = None

    def update_autolog_state(
        self,
    ):
        # Currently we disable autologging for better control in AutoML
        _autolog = is_autolog_enabled()
        self._do_log_model = AUTOLOGGING_INTEGRATIONS["mlflow"].get("log_models", True)
        if self.experiment_type == "automl":
            self.autolog = False
            self.manual_log = mlflow.active_run() is not None or _autolog
            self.log_type = "manual"
            if _autolog:
                logger.debug("Disabling autologging")
                self.resume_params = AUTOLOGGING_INTEGRATIONS["mlflow"].copy()
                mlflow.autolog(disable=True, silent=True, log_models=self._do_log_model)
                self.log_type = "r_autolog"  # 'r' for replace autolog with manual log

        elif self.experiment_type == "tune":
            self.autolog = _autolog
            self.manual_log = not self.autolog and mlflow.active_run() is not None

            if self.autolog:
                self.log_type = "autolog"

            if self.manual_log:
                self.log_type = "manual"
        else:
            raise ValueError(f"Unknown experiment type: {self.experiment_type}")

    def copy_mlflow_run(self, src_id, target_id, components=["param", "metric", "tag"]):
        src_run = self.mlflow_client.get_run(src_id)
        if "param" in components:
            for param_name, param_value in src_run.data.params.items():
                try:
                    self.mlflow_client.log_param(target_id, param_name, param_value)
                except mlflow.exceptions.MlflowException:
                    pass

        timestamp = int(time.time() * 1000)

        if "metric" in components:
            _metrics = [Metric(key, value, timestamp, 0) for key, value in src_run.data.metrics.items()]
        else:
            _metrics = []

        if "tag" in components:
            _tags = [
                RunTag(key, str(value))
                for key, value in src_run.data.tags.items()
                if key.startswith("flaml") or key.startswith("synapseml")
            ]
        else:
            _tags = []
        self.mlflow_client.log_batch(run_id=target_id, metrics=_metrics, params=[], tags=_tags)
        return f"Successfully copy_mlflow_run run_id {src_id} to run_id {target_id}"

    def record_trial(self, result, trial, metric):
        if isinstance(result, dict):
            metrics = flatten_dict(result)
            metric_name = str(list(metrics.keys()))
        else:
            metrics = {metric: result}
            metric_name = metric

        if "ml" in trial.config.keys():
            params = trial.config["ml"]
        else:
            params = trial.config

        info = {
            "metrics": metrics,
            "params": params,
            "tags": {
                "flaml.best_run": False,
                "flaml.iteration_number": self.child_counter,
                "flaml.version": __version__,
                "flaml.meric": metric_name,
                "flaml.run_source": "flaml-tune",
                "flaml.log_type": self.log_type,
            },
            "submetrics": {
                "values": [],
            },
        }

        self.infos.append(info)

        if not self.autolog and not self.manual_log:
            return

        if self.manual_log:
            with mlflow.start_run(
                nested=True, run_name=f"{self.parent_run_name}_child_{self.child_counter}"
            ) as child_run:
                self._log_info_to_run(info, child_run.info.run_id, log_params=True)
                self.manual_run_ids.append(child_run.info.run_id)
        self.child_counter += 1

    def log_tune(self, analysis, metric):
        self.set_best_iter(analysis)
        if self.autolog:
            if self.parent_run_id is not None:
                mlflow.start_run(run_id=self.parent_run_id, experiment_id=self.experiment_id)
                mlflow.log_metric("num_child_runs", len(self.infos))
            self.adopt_children(analysis)

        if self.manual_log:
            if "ml" in analysis.best_config.keys():
                mlflow.log_params(analysis.best_config["ml"])
            else:
                mlflow.log_params(analysis.best_config)
            mlflow.log_metric("best_" + metric, analysis.best_result[metric])
            best_mlflow_run_id = self.manual_run_ids[analysis.best_iteration]
            best_mlflow_run_name = self.mlflow_client.get_run(best_mlflow_run_id).info.run_name
            analysis.best_run_id = best_mlflow_run_id
            analysis.best_run_name = best_mlflow_run_name
            self.mlflow_client.set_tag(best_mlflow_run_id, "flaml.best_run", True)
            self.best_run_id = best_mlflow_run_id
            if not self.has_summary:
                self.copy_mlflow_run(best_mlflow_run_id, self.parent_run_id)
                self.has_summary = True

    def log_model(self, model, estimator, signature=None, run_id=None):
        if not self._do_log_model:
            return
        logger.debug(f"logging model {estimator}")
        ret_message = f"Successfully log_model {estimator} to run_id {run_id}"
        optional_remove_list = (
            [] if estimator in ["transformer", "transformer_ms", "tcn", "tft"] else OPTIONAL_REMOVE_REQUIREMENT_LIST
        )
        run = mlflow.active_run()
        if run and run.info.run_id == self.parent_run_id:
            logger.debug(
                f"Current active run_id {run.info.run_id} == parent_run_id {self.parent_run_id}, Starting run_id {run_id}"
            )
            mlflow.start_run(run_id=run_id, nested=True)
        elif run and run.info.run_id != run_id:
            ret_message = (
                f"Error: Should log_model {estimator} to run_id {run_id}, but logged to run_id {run.info.run_id}"
            )
            logger.error(ret_message)
        else:
            logger.debug(f"No active run, start run_id {run_id}")
            mlflow.start_run(run_id=run_id)
        logger.debug(f"logged model {estimator} to run_id {mlflow.active_run().info.run_id}")
        if estimator.endswith("_spark"):
            # mlflow.spark.log_model(model, estimator, signature=signature)
            mlflow.spark.log_model(model, "model", signature=signature)
        elif estimator in ["lgbm"]:
            mlflow.lightgbm.log_model(model, estimator, signature=signature)
        elif estimator in ["transformer", "transformer_ms"]:
            mlflow.transformers.log_model(model, estimator, signature=signature)
        elif estimator in ["arima", "sarimax", "holt-winters", "snaive", "naive", "savg", "avg", "ets"]:
            mlflow.statsmodels.log_model(model, estimator, signature=signature)
        elif estimator in ["tcn", "tft"]:
            mlflow.pytorch.log_model(model, estimator, signature=signature)
        elif estimator in ["prophet"]:
            mlflow.prophet.log_model(model, estimator, signature=signature)
        elif estimator in ["orbit"]:
            logger.warning(f"Unsupported model: {estimator}. No model logged.")
        else:
            mlflow.sklearn.log_model(model, estimator, signature=signature)
        future = executor.submit(
            lambda: mlflow.models.model.update_model_requirements(
                model_uri=f"runs:/{run_id}/{'model' if estimator.endswith('_spark') else estimator}",
                operation="remove",
                requirement_list=convert_requirement(REMOVE_REQUIREMENT_LIST + optional_remove_list),
            )
        )
        self.futures[future] = f"run_{run_id}_requirements_updated"
        if not run or run.info.run_id == self.parent_run_id:
            logger.debug(f"Ending current run_id {mlflow.active_run().info.run_id}")
            mlflow.end_run()
        return ret_message

    def _pickle_and_log_artifact(self, obj, artifact_name, pickle_fname="temp_.pkl", run_id=None):
        if not self._do_log_model:
            return True
        with tempfile.TemporaryDirectory() as tmpdir:
            pickle_fpath = os.path.join(tmpdir, pickle_fname)
            try:
                with open(pickle_fpath, "wb") as f:
                    pickle.dump(obj, f)
                mlflow.log_artifact(pickle_fpath, artifact_name, run_id)
                return True
            except Exception as e:
                logger.debug(f"Failed to pickle and log {artifact_name}, error: {e}")
                return False

    def _log_pipeline(self, pipeline, flavor_name, pipeline_name, signature, run_id, estimator=None):
        logger.debug(f"logging pipeline {flavor_name}:{pipeline_name}:{estimator}")
        ret_message = f"Successfully _log_pipeline {flavor_name}:{pipeline_name}:{estimator} to run_id {run_id}"
        optional_remove_list = (
            [] if estimator in ["transformer", "transformer_ms", "tcn", "tft"] else OPTIONAL_REMOVE_REQUIREMENT_LIST
        )
        run = mlflow.active_run()
        if run and run.info.run_id == self.parent_run_id:
            logger.debug(
                f"Current active run_id {run.info.run_id} == parent_run_id {self.parent_run_id}, Starting run_id {run_id}"
            )
            mlflow.start_run(run_id=run_id, nested=True)
        elif run and run.info.run_id != run_id:
            ret_message = f"Error: Should _log_pipeline {flavor_name}:{pipeline_name}:{estimator} model to run_id {run_id}, but logged to run_id {run.info.run_id}"
            logger.error(ret_message)
        else:
            logger.debug(f"No active run, start run_id {run_id}")
            mlflow.start_run(run_id=run_id)
        logger.debug(
            f"logging pipeline {flavor_name}:{pipeline_name}:{estimator} to run_id {mlflow.active_run().info.run_id}"
        )
        if flavor_name == "sklearn":
            mlflow.sklearn.log_model(pipeline, pipeline_name, signature=signature)
        elif flavor_name == "spark":
            mlflow.spark.log_model(pipeline, pipeline_name, signature=signature)
        else:
            logger.warning(f"Unsupported pipeline flavor: {flavor_name}. No model logged.")
        future = executor.submit(
            lambda: mlflow.models.model.update_model_requirements(
                model_uri=f"runs:/{run_id}/{pipeline_name}",
                operation="remove",
                requirement_list=convert_requirement(REMOVE_REQUIREMENT_LIST + optional_remove_list),
            )
        )
        self.futures[future] = f"run_{run_id}_requirements_updated"
        if not run or run.info.run_id == self.parent_run_id:
            logger.debug(f"Ending current run_id {mlflow.active_run().info.run_id}")
            mlflow.end_run()
        return ret_message

    def pickle_and_log_automl_artifacts(self, automl, model, estimator, signature=None, run_id=None):
        """log automl artifacts to mlflow
        load back with `automl = mlflow.pyfunc.load_model(model_run_id_or_uri)`, then do prediction with `automl.predict(X)`
        """
        logger.debug(f"logging automl estimator {estimator}")
        # self._pickle_and_log_artifact(
        #     automl.feature_transformer, "feature_transformer", "feature_transformer.pkl", run_id
        # )
        # self._pickle_and_log_artifact(automl.label_transformer, "label_transformer", "label_transformer.pkl", run_id)
        if estimator.endswith("_spark"):
            # spark pipeline is not supported yet
            return
        feature_transformer = automl.feature_transformer
        if isinstance(feature_transformer, Pipeline) and not estimator.endswith("_spark"):
            pipeline = feature_transformer
            pipeline.steps.append(("estimator", model))
        elif isinstance(feature_transformer, SparkPipelineModel) and estimator.endswith("_spark"):
            pipeline = feature_transformer
            pipeline.stages.append(model)
        elif not estimator.endswith("_spark"):
            steps = [("feature_transformer", feature_transformer)]
            steps.append(("estimator", model))
            pipeline = Pipeline(steps)
        else:
            stages = []
            if feature_transformer is not None:
                stages.append(feature_transformer)
            stages.append(model)
            pipeline = SparkPipelineModel(stages=stages)
        if isinstance(pipeline, SparkPipelineModel):
            logger.debug(f"logging spark pipeline {estimator}")
            self._log_pipeline(pipeline, "spark", "model", signature, run_id, estimator)
        else:
            # Add a log named "model" to fit default settings
            logger.debug(f"logging sklearn pipeline {estimator}")
            self._log_pipeline(pipeline, "sklearn", "model", signature, run_id, estimator)
        return f"Successfully pickle_and_log_automl_artifacts {estimator} to run_id {run_id}"

    @time_it
    def record_state(self, automl, search_state, estimator):
        _st = time.time()
        automl_metric_name = (
            automl._state.metric if isinstance(automl._state.metric, str) else automl._state.error_metric
        )
        if automl._state.error_metric.startswith("1-"):
            automl_metric_value = 1 - search_state.val_loss
        elif automl._state.error_metric.startswith("-"):
            automl_metric_value = -search_state.val_loss
        else:
            automl_metric_value = search_state.val_loss

        if "ml" in search_state.config:
            config = search_state.config["ml"]
        else:
            config = search_state.config

        self.automl_user_configurations = safe_json_dumps(automl._automl_user_configurations)

        info = {
            "metrics": {
                "iter_counter": automl._track_iter,
                "trial_time": search_state.trial_time,
                "wall_clock_time": automl._state.time_from_start,
                "validation_loss": search_state.val_loss,
                "best_validation_loss": search_state.best_loss,
                automl_metric_name: automl_metric_value,
            },
            "tags": {
                "flaml.best_run": False,
                "flaml.estimator_name": estimator,
                "flaml.estimator_class": search_state.learner_class.__name__,
                "flaml.iteration_number": automl._track_iter,
                "flaml.version": __version__,
                "flaml.learner": estimator,
                "flaml.sample_size": search_state.sample_size,
                "flaml.meric": automl_metric_name,
                "flaml.run_source": "flaml-automl",
                "flaml.log_type": self.log_type,
                "flaml.automl_user_configurations": self.automl_user_configurations,
            },
            "params": {
                "sample_size": search_state.sample_size,
                "learner": estimator,
                **config,
            },
            "submetrics": {
                "iter_counter": automl._iter_per_learner[estimator],
                "values": [],
            },
        }

        if (search_state.metric_for_logging is not None) and (
            "intermediate_results" in search_state.metric_for_logging
        ):
            info["submetrics"]["values"] = search_state.metric_for_logging["intermediate_results"]

        self.infos.append(info)

        if not self.autolog and not self.manual_log:
            return
        if self.manual_log:
            if self.parent_run_name is not None:
                run_name = f"{self.parent_run_name}_child_{self.child_counter}"
            else:
                run_name = None
            _t1 = time.time()
            wait(self.futures_log_model)
            _t2 = time.time() - _t1
            logger.debug(f"wait futures_log_model in record_state took {_t2} seconds")
            with mlflow.start_run(nested=True, run_name=run_name) as child_run:
                future = executor.submit(lambda: self._log_info_to_run(info, child_run.info.run_id, log_params=True))
                self.futures[future] = f"iter_{automl._track_iter}_log_info_to_run"
                future = executor.submit(lambda: self._log_automl_configurations(child_run.info.run_id))
                self.futures[future] = f"iter_{automl._track_iter}_log_automl_configurations"
                if automl._state.model_history:
                    if estimator.endswith("_spark"):
                        future = executor.submit(
                            lambda: self.log_model(
                                search_state.trained_estimator._model,
                                estimator,
                                automl.estimator_signature,
                                child_run.info.run_id,
                            )
                        )
                        self.futures_log_model[future] = f"record_state-log_model_{estimator}"
                    else:
                        future = executor.submit(
                            lambda: self.pickle_and_log_automl_artifacts(
                                automl,
                                search_state.trained_estimator,
                                estimator,
                                automl.pipeline_signature,
                                child_run.info.run_id,
                            )
                        )
                        self.futures_log_model[future] = f"record_state-pickle_and_log_automl_artifacts_{estimator}"
                self.manual_run_ids.append(child_run.info.run_id)
            self.child_counter += 1
        return f"Successfully record_state iteration {automl._track_iter}"

    @time_it
    def log_automl(self, automl):
        self.set_best_iter(automl)
        if self.autolog:
            if self.parent_run_id is not None:
                mlflow.start_run(run_id=self.parent_run_id, experiment_id=self.experiment_id)
                mlflow.log_metrics(
                    {
                        "best_validation_loss": automl._state.best_loss,
                        "best_iteration": automl._best_iteration,
                        "num_child_runs": len(self.infos),
                    }
                )
                if (
                    automl._trained_estimator is not None
                    and not self.has_model
                    and automl._trained_estimator._model is not None
                ):
                    if automl.best_estimator.endswith("_spark"):
                        self.log_model(
                            automl._trained_estimator._model,
                            automl.best_estimator,
                            automl.estimator_signature,
                            self.parent_run_id,
                        )
                    else:
                        self.pickle_and_log_automl_artifacts(
                            automl, automl.model, automl.best_estimator, automl.pipeline_signature, self.parent_run_id
                        )
                    self.has_model = True

            self.adopt_children(automl)

        if self.manual_log:
            best_mlflow_run_id = self.manual_run_ids[automl._best_iteration]
            best_run_name = self.mlflow_client.get_run(best_mlflow_run_id).info.run_name
            automl.best_run_id = best_mlflow_run_id
            automl.best_run_name = best_run_name
            self.mlflow_client.set_tag(best_mlflow_run_id, "flaml.best_run", True)
            self.best_run_id = best_mlflow_run_id
            if self.parent_run_id is not None:
                conf = automl._config_history[automl._best_iteration][1].copy()
                if "ml" in conf.keys():
                    conf = conf["ml"]

                mlflow.log_params({**conf, "best_learner": automl._best_estimator}, run_id=self.parent_run_id)
                if not self.has_summary:
                    logger.info(f"logging best model {automl.best_estimator}")
                    future = executor.submit(lambda: self.copy_mlflow_run(best_mlflow_run_id, self.parent_run_id))
                    self.futures[future] = "log_automl_copy_mlflow_run"
                    future = executor.submit(lambda: self._log_automl_configurations(self.parent_run_id))
                    self.futures[future] = "log_automl_log_automl_configurations"
                    self.has_summary = True
                    _t1 = time.time()
                    wait(self.futures_log_model)
                    _t2 = time.time() - _t1
                    logger.debug(f"wait futures_log_model in log_automl took {_t2} seconds")
                    if (
                        automl._trained_estimator is not None
                        and not self.has_model
                        and automl._trained_estimator._model is not None
                    ):
                        if automl.best_estimator.endswith("_spark"):
                            future = executor.submit(
                                lambda: self.log_model(
                                    automl._trained_estimator._model,
                                    automl.best_estimator,
                                    signature=automl.estimator_signature,
                                    run_id=self.parent_run_id,
                                )
                            )
                            self.futures_log_model[future] = f"log_automl-log_model_{automl.best_estimator}"
                        else:
                            future = executor.submit(
                                lambda: self.pickle_and_log_automl_artifacts(
                                    automl,
                                    automl.model,
                                    automl.best_estimator,
                                    signature=automl.pipeline_signature,
                                    run_id=self.parent_run_id,
                                )
                            )
                            self.futures_log_model[
                                future
                            ] = f"log_automl-pickle_and_log_automl_artifacts_{automl.best_estimator}"
                        self.has_model = True

    def resume_mlflow(self):
        if len(self.resume_params) > 0:
            mlflow.autolog(**self.resume_params)

    def _log_automl_configurations(self, run_id):
        self.mlflow_client.log_text(
            run_id=run_id,
            text=self.automl_user_configurations,
            artifact_file="automl_configurations/automl_user_configurations.json",
        )
        return f"Successfully _log_automl_configurations to run_id {run_id}"

    def _log_info_to_run(self, info, run_id, log_params=False):
        _metrics = [Metric(key, value, int(time.time() * 1000), 0) for key, value in info["metrics"].items()]
        _tags = [
            RunTag(key, str(value)[:5000]) for key, value in info["tags"].items()
        ]  # AML will raise error if value length > 5000
        _params = [
            Param(key, str(value))
            for key, value in info["params"].items()
            if log_params or key in ["sample_size", "learner"]
        ]
        self.mlflow_client.log_batch(run_id=run_id, metrics=_metrics, params=_params, tags=_tags)

        if len(info["submetrics"]["values"]) > 0:
            for each_entry in info["submetrics"]["values"]:
                with mlflow.start_run(nested=True) as run:
                    each_entry.update({"iter_counter": info["submetrics"]["iter_counter"]})
                    _metrics = [Metric(key, value, int(time.time() * 1000), 0) for key, value in each_entry.items()]
                    _tags = [RunTag("mlflow.parentRunId", run_id)]
                    self.mlflow_client.log_batch(run_id=run.info.run_id, metrics=_metrics, params=[], tags=_tags)
            del info["submetrics"]["values"]
        return f"Successfully _log_info_to_run to run_id {run_id}"

    def adopt_children(self, result=None):
        """
        Set autologging child runs to nested by fetching them after all child runs are completed.
        Note that this may cause disorder when concurrently starting multiple AutoML processes
        with the same experiment name if the MLflow version is less than or equal to "2.5.0".
        """
        if self.autolog:
            best_iteration = self.best_iteration
            if best_iteration is None:
                logger.warning("best_iteration is None, cannot identify best run")
            raw_autolog_child_runs = mlflow.search_runs(
                experiment_ids=[self.experiment_id],
                order_by=["attributes.start_time DESC"],
                max_results=SEARCH_MAX_RESULTS,
                output_format="list",
                filter_string=(
                    f"tags.extra_tag.sid = '{self.extra_tag['extra_tag.sid']}'" if mlflow.__version__ > "2.5.0" else ""
                ),
            )
            self.child_counter = 0

            # From latest to earliest, remove duplicate cross-validation runs
            _exist_child_run_params = []  # for deduplication of cross-validation child runs
            _to_keep_autolog_child_runs = []
            for autolog_child_run in raw_autolog_child_runs:
                child_start_time = autolog_child_run.info.start_time / 1000

                if child_start_time < self.start_time:
                    continue

                _current_child_run_params = autolog_child_run.data.params
                # remove n_estimators as some models will train with small n_estimators to estimate time budget
                if self.experiment_type == "automl":
                    _current_child_run_params.pop("n_estimators", None)
                if _current_child_run_params in _exist_child_run_params:
                    # remove duplicate cross-validation run
                    self.mlflow_client.delete_run(autolog_child_run.info.run_id)
                    continue
                else:
                    _exist_child_run_params.append(_current_child_run_params)
                    _to_keep_autolog_child_runs.append(autolog_child_run)

            # From earliest to latest, set tags and child_counter
            autolog_child_runs = _to_keep_autolog_child_runs[::-1]
            for autolog_child_run in autolog_child_runs:
                child_run_id = autolog_child_run.info.run_id
                child_run_parent_id = autolog_child_run.data.tags.get("mlflow.parentRunId", None)
                child_start_time = autolog_child_run.info.start_time / 1000

                if child_start_time < self.start_time:
                    continue

                if all(
                    [
                        len(autolog_child_run.data.params) == 0,
                        len(autolog_child_run.data.metrics) == 0,
                        child_run_id != self.parent_run_id,
                    ]
                ):
                    # remove empty run
                    # empty run could be created by mlflow autologging
                    self.mlflow_client.delete_run(autolog_child_run.info.run_id)
                    continue

                if all(
                    [
                        child_run_id != self.parent_run_id,
                        child_run_parent_id is None or child_run_parent_id == self.parent_run_id,
                    ]
                ):
                    if self.parent_run_id is not None:
                        self.mlflow_client.set_tag(
                            child_run_id,
                            "mlflow.parentRunId",
                            self.parent_run_id,
                        )
                        if IS_RENAME_CHILD_RUN:
                            self.mlflow_client.set_tag(
                                child_run_id,
                                "mlflow.runName",
                                f"{self.parent_run_name}_child_{self.child_counter}",
                            )
                        self.mlflow_client.set_tag(child_run_id, "flaml.child_counter", self.child_counter)

                    # merge autolog child run and corresponding manual run
                    flaml_info = self.infos[self.child_counter]
                    child_run = self.mlflow_client.get_run(child_run_id)
                    self._log_info_to_run(flaml_info, child_run_id, log_params=False)

                    if self.experiment_type == "automl":
                        if "learner" not in child_run.data.params:
                            self.mlflow_client.log_param(child_run_id, "learner", flaml_info["params"]["learner"])
                        if "sample_size" not in child_run.data.params:
                            self.mlflow_client.log_param(
                                child_run_id, "sample_size", flaml_info["params"]["sample_size"]
                            )

                    if self.child_counter == best_iteration:
                        self.mlflow_client.set_tag(child_run_id, "flaml.best_run", True)
                        if result is not None:
                            result.best_run_id = child_run_id
                            result.best_run_name = child_run.info.run_name
                            self.best_run_id = child_run_id
                        if self.parent_run_id is not None and not self.has_summary:
                            self.copy_mlflow_run(child_run_id, self.parent_run_id)
                            self.has_summary = True
                    self.child_counter += 1

    def retrain(self, train_func, config):
        """retrain with given config, added for logging the best config and model to parent run.
        No more needed after v2.0.2post2 as we no longer log best config and model to parent run.
        """
        if self.autolog:
            self.set_mlflow_config()
            self.has_summary = True
            with mlflow.start_run(run_id=self.parent_run_id):
                train_func(config)

    def __del__(self):
        # mlflow.end_run()  # this will end the parent run when re-fit an AutoML instance. Bug 2922020: Inconsistent Run Creation Output
        self.resume_mlflow()


def register_automl_pipeline(automl, model_name=None, signature=None):
    pipeline = automl.automl_pipeline
    if pipeline is None:
        logger.warning("pipeline not found, cannot register it")
        return
    if model_name is None:
        model_name = automl._mlflow_exp_name + "_pipeline"
    if automl.best_run_id is None:
        mlflow.sklearn.log_model(
            pipeline,
            "automl_pipeline",
            registered_model_name=model_name,
            signature=automl.pipeline_signature if signature is None else signature,
        )
        mvs = mlflow.search_model_versions(
            filter_string=f"name='{model_name}'", order_by=["attribute.version_number ASC"], max_results=1
        )
        return mvs[0]
    else:
        best_run = mlflow.get_run(automl.best_run_id)
        model_uri = f"runs:/{best_run.info.run_id}/automl_pipeline"
        return mlflow.register_model(model_uri, model_name)



# Copyright 2020 The Ray Authors.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This source file is adapted here because ray does not fully support Windows.

# Copyright (c) Microsoft Corporation.
import copy
import logging
import random
from typing import Any, Dict, Generator, List, Tuple

import numpy

from ..sample import Categorical, Domain, RandomState

try:
    from ray import __version__ as ray_version

    if ray_version.startswith("1."):
        from ray.tune.sample import Domain as RayDomain
    else:
        from ray.tune.search.sample import Domain as RayDomain
except ImportError:
    RayDomain = Domain

logger = logging.getLogger(__name__)


class TuneError(Exception):
    """General error class raised by ray.tune."""

    pass


def generate_variants(
    unresolved_spec: Dict,
    constant_grid_search: bool = False,
    random_state: "RandomState" = None,
) -> Generator[Tuple[Dict, Dict], None, None]:
    """Generates variants from a spec (dict) with unresolved values.
    There are two types of unresolved values:
        Grid search: These define a grid search over values. For example, the
        following grid search values in a spec will produce six distinct
        variants in combination:
            "activation": grid_search(["relu", "tanh"])
            "learning_rate": grid_search([1e-3, 1e-4, 1e-5])
        Lambda functions: These are evaluated to produce a concrete value, and
        can express dependencies or conditional distributions between values.
        They can also be used to express random search (e.g., by calling
        into the `random` or `np` module).
            "cpu": lambda spec: spec.config.num_workers
            "batch_size": lambda spec: random.uniform(1, 1000)
    Finally, to support defining specs in plain JSON / YAML, grid search
    and lambda functions can also be defined alternatively as follows:
        "activation": {"grid_search": ["relu", "tanh"]}
        "cpu": {"eval": "spec.config.num_workers"}
    Use `format_vars` to format the returned dict of hyperparameters.
    Yields:
        (Dict of resolved variables, Spec object)
    """
    for resolved_vars, spec in _generate_variants(
        unresolved_spec,
        constant_grid_search=constant_grid_search,
        random_state=random_state,
    ):
        assert not _unresolved_values(spec)
        yield resolved_vars, spec


def grid_search(values: List) -> Dict[str, List]:
    """Convenience method for specifying grid search over a value.
    Arguments:
        values: An iterable whose parameters will be gridded.
    """

    return {"grid_search": values}


_STANDARD_IMPORTS = {
    "random": random,
    "np": numpy,
}

_MAX_RESOLUTION_PASSES = 20


def parse_spec_vars(
    spec: Dict,
) -> Tuple[List[Tuple[Tuple, Any]], List[Tuple[Tuple, Any]], List[Tuple[Tuple, Any]]]:
    resolved, unresolved = _split_resolved_unresolved_values(spec)
    resolved_vars = list(resolved.items())

    if not unresolved:
        return resolved_vars, [], []

    grid_vars = []
    domain_vars = []
    for path, value in unresolved.items():
        if value.is_grid():
            grid_vars.append((path, value))
        else:
            domain_vars.append((path, value))
    grid_vars.sort()

    return resolved_vars, domain_vars, grid_vars


def _generate_variants(
    spec: Dict, constant_grid_search: bool = False, random_state: "RandomState" = None
) -> Tuple[Dict, Dict]:
    spec = copy.deepcopy(spec)
    _, domain_vars, grid_vars = parse_spec_vars(spec)

    if not domain_vars and not grid_vars:
        yield {}, spec
        return

    # Variables to resolve
    to_resolve = domain_vars

    all_resolved = True
    if constant_grid_search:
        # In this path, we first sample random variables and keep them constant
        # for grid search.
        # `_resolve_domain_vars` will alter `spec` directly
        all_resolved, resolved_vars = _resolve_domain_vars(
            spec, domain_vars, allow_fail=True, random_state=random_state
        )
        if not all_resolved:
            # Not all variables have been resolved, but remove those that have
            # from the `to_resolve` list.
            to_resolve = [(r, d) for r, d in to_resolve if r not in resolved_vars]
    grid_search = _grid_search_generator(spec, grid_vars)
    for resolved_spec in grid_search:
        if not constant_grid_search or not all_resolved:
            # In this path, we sample the remaining random variables
            _, resolved_vars = _resolve_domain_vars(resolved_spec, to_resolve, random_state=random_state)

        for resolved, spec in _generate_variants(
            resolved_spec,
            constant_grid_search=constant_grid_search,
            random_state=random_state,
        ):
            for path, value in grid_vars:
                resolved_vars[path] = _get_value(spec, path)
            for k, v in resolved.items():
                if k in resolved_vars and v != resolved_vars[k] and _is_resolved(resolved_vars[k]):
                    raise ValueError(
                        "The variable `{}` could not be unambiguously "
                        "resolved to a single value. Consider simplifying "
                        "your configuration.".format(k)
                    )
                resolved_vars[k] = v
            yield resolved_vars, spec


def assign_value(spec: Dict, path: Tuple, value: Any):
    for k in path[:-1]:
        spec = spec[k]
    spec[path[-1]] = value


def _get_value(spec: Dict, path: Tuple) -> Any:
    for k in path:
        spec = spec[k]
    return spec


def _resolve_domain_vars(
    spec: Dict,
    domain_vars: List[Tuple[Tuple, Domain]],
    allow_fail: bool = False,
    random_state: "RandomState" = None,
) -> Tuple[bool, Dict]:
    resolved = {}
    error = True
    num_passes = 0
    while error and num_passes < _MAX_RESOLUTION_PASSES:
        num_passes += 1
        error = False
        for path, domain in domain_vars:
            if path in resolved:
                continue
            try:
                value = domain.sample(_UnresolvedAccessGuard(spec), random_state=random_state)
            except RecursiveDependencyError as e:
                error = e
            # except Exception:
            #     raise ValueError(
            #         "Failed to evaluate expression: {}: {}".format(path, domain)
            #     )
            else:
                assign_value(spec, path, value)
                resolved[path] = value
    if error:
        if not allow_fail:
            raise error
        else:
            return False, resolved
    return True, resolved


def _grid_search_generator(unresolved_spec: Dict, grid_vars: List) -> Generator[Dict, None, None]:
    value_indices = [0] * len(grid_vars)

    def increment(i):
        value_indices[i] += 1
        if value_indices[i] >= len(grid_vars[i][1]):
            value_indices[i] = 0
            if i + 1 < len(value_indices):
                return increment(i + 1)
            else:
                return True
        return False

    if not grid_vars:
        yield unresolved_spec
        return

    while value_indices[-1] < len(grid_vars[-1][1]):
        spec = copy.deepcopy(unresolved_spec)
        for i, (path, values) in enumerate(grid_vars):
            assign_value(spec, path, values[value_indices[i]])
        yield spec
        if grid_vars:
            done = increment(0)
            if done:
                break


def _is_resolved(v) -> bool:
    resolved, _ = _try_resolve(v)
    return resolved


def _try_resolve(v) -> Tuple[bool, Any]:
    if isinstance(v, (Domain, RayDomain)):
        # Domain to sample from
        return False, v
    elif isinstance(v, dict) and len(v) == 1 and "grid_search" in v:
        # Grid search values
        grid_values = v["grid_search"]
        if not isinstance(grid_values, list):
            raise TuneError(f"Grid search expected list of values, got: {grid_values}")
        return False, Categorical(grid_values).grid()
    return True, v


def _split_resolved_unresolved_values(
    spec: Dict,
) -> Tuple[Dict[Tuple, Any], Dict[Tuple, Any]]:
    resolved_vars = {}
    unresolved_vars = {}
    for k, v in spec.items():
        resolved, v = _try_resolve(v)
        if not resolved:
            unresolved_vars[(k,)] = v
        elif isinstance(v, dict):
            # Recurse into a dict
            (
                _resolved_children,
                _unresolved_children,
            ) = _split_resolved_unresolved_values(v)
            for path, value in _resolved_children.items():
                resolved_vars[(k,) + path] = value
            for path, value in _unresolved_children.items():
                unresolved_vars[(k,) + path] = value
        elif isinstance(v, list):
            # Recurse into a list
            for i, elem in enumerate(v):
                (
                    _resolved_children,
                    _unresolved_children,
                ) = _split_resolved_unresolved_values({i: elem})
                for path, value in _resolved_children.items():
                    resolved_vars[(k,) + path] = value
                for path, value in _unresolved_children.items():
                    unresolved_vars[(k,) + path] = value
        else:
            resolved_vars[(k,)] = v
    return resolved_vars, unresolved_vars


def _unresolved_values(spec: Dict) -> Dict[Tuple, Any]:
    return _split_resolved_unresolved_values(spec)[1]


def has_unresolved_values(spec: Dict) -> bool:
    return True if _unresolved_values(spec) else False


class _UnresolvedAccessGuard(dict):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.__dict__ = self

    def __getattribute__(self, item):
        value = dict.__getattribute__(self, item)
        if not _is_resolved(value):
            raise RecursiveDependencyError(f"`{item}` recursively depends on {value}")
        elif isinstance(value, dict):
            return _UnresolvedAccessGuard(value)
        else:
            return value


class RecursiveDependencyError(Exception):
    def __init__(self, msg: str):
        Exception.__init__(self, msg)



import datetime
import math
from functools import lru_cache

import pandas as pd


def monthly_fourier_features(timestamps: pd.Series, month_fourier_degree: int = 2):
    if len(timestamps):
        data = pd.DataFrame({"time": timestamps})
        month_pos = timestamps.apply(lambda x: position_in_month(datetime.date(x.year, x.month, x.day)))
        for d in range(month_fourier_degree):
            data[f"cos{d+1}"] = (2 * (d + 1) * math.pi * month_pos).apply(math.cos)
            data[f"sin{d + 1}"] = (2 * (d + 1) * math.pi * month_pos).apply(math.sin)

        drop_cols = ["time"]
        data = data.drop(columns=drop_cols)
        return data
    else:
        columns = []
        for d in range(month_fourier_degree):
            columns += [f"cos{d+1}", f"sin{d + 1}"]

        return pd.DataFrame(columns=columns)


@lru_cache(maxsize=4096)
def position_in_month(d: datetime.date):
    prev = datetime.date(d.year, d.month, 1) - datetime.timedelta(days=1)
    nxt = datetime.date(
        d.year + 1 if d.month == 12 else d.year, 1 if d.month == 12 else d.month + 1, 1
    ) - datetime.timedelta(days=1)
    delta = (d - prev).days / (nxt - prev).days
    return delta



import logging
import os


class ColoredFormatter(logging.Formatter):
    # ANSI escape codes for colors
    COLORS = {
        # logging.DEBUG: "\033[36m",  # Cyan
        # logging.INFO: "\033[32m",   # Green
        logging.WARNING: "\033[33m",  # Yellow
        logging.ERROR: "\033[31m",  # Red
        logging.CRITICAL: "\033[1;31m",  # Bright Red
    }
    RESET = "\033[0m"  # Reset to default

    def __init__(self, fmt, datefmt, use_color=True):
        super().__init__(fmt, datefmt)
        self.use_color = use_color

    def format(self, record):
        formatted = super().format(record)
        if self.use_color:
            color = self.COLORS.get(record.levelno, "")
            if color:
                return f"{color}{formatted}{self.RESET}"
        return formatted


logger = logging.getLogger(__name__)
use_color = True
if os.getenv("FLAML_LOG_NO_COLOR"):
    use_color = False

logger_formatter = ColoredFormatter(
    "[%(name)s: %(asctime)s] {%(lineno)d} %(levelname)s - %(message)s", "%m-%d %H:%M:%S", use_color
)
logger.propagate = False



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import numpy as np

from flaml.automl.data import DataFrame, Series, psDataFrame, psSeries

if TYPE_CHECKING:
    import flaml

# TODO: if your task is not specified in here, define your task as an all-capitalized word
SEQCLASSIFICATION = "seq-classification"
MULTICHOICECLASSIFICATION = "multichoice-classification"
TOKENCLASSIFICATION = "token-classification"

SEQREGRESSION = "seq-regression"

TS_FORECASTREGRESSION = (
    "forecast",
    "ts_forecast",
    "ts_forecast_regression",
)
REGRESSION = ("regression", SEQREGRESSION, *TS_FORECASTREGRESSION)
TS_FORECASTCLASSIFICATION = "ts_forecast_classification"
TS_FORECASTPANEL = "ts_forecast_panel"
TS_FORECAST = (
    *TS_FORECASTREGRESSION,
    TS_FORECASTCLASSIFICATION,
    TS_FORECASTPANEL,
)
CLASSIFICATION = (
    "binary",
    "multiclass",
    "classification",
    SEQCLASSIFICATION,
    MULTICHOICECLASSIFICATION,
    TOKENCLASSIFICATION,
    TS_FORECASTCLASSIFICATION,
)
RANK = ("rank",)
SUMMARIZATION = "summarization"
NLG_TASKS = (SUMMARIZATION,)
NLU_TASKS = (
    SEQREGRESSION,
    SEQCLASSIFICATION,
    MULTICHOICECLASSIFICATION,
    TOKENCLASSIFICATION,
)
NLP_TASKS = (*NLG_TASKS, *NLU_TASKS)


def get_classification_objective(num_labels: int) -> str:
    if num_labels == 2:
        objective_name = "binary"
    else:
        objective_name = "multiclass"
    return objective_name


class Task(ABC):
    """
    Abstract base class for a machine learning task.

    Class definitions should implement abstract methods and provide a non-empty dictionary of estimator classes.
    A Task can be suitable to be used for multiple machine-learning tasks (e.g. classification or regression) or be
    implemented specifically for a single one depending on the generality of data validation and model evaluation methods
    implemented. The implementation of a Task may optionally use the training data and labels to determine data and task
    specific details, such as in determining if a problem is single-label or multi-label.

    FLAML evaluates at runtime how to behave exactly, relying on the task instance to provide implementations of
    operations which vary between tasks.
    """

    def __init__(
        self,
        task_name: str,
        X_train: Optional[Union[np.ndarray, DataFrame, psDataFrame]] = None,
        y_train: Optional[Union[np.ndarray, DataFrame, Series, psSeries]] = None,
    ):
        """Constructor.

        Args:
            task_name: String name for this type of task. Used when the Task can be generic and implement a number of
                types of sub-task.
            X_train: Optional. Some Task types may use the data shape or features to determine details of their usage,
                such as in binary vs multilabel classification.
            y_train: Optional. Some Task types may use the data shape or features to determine details of their usage,
                such as in binary vs multilabel classification.
        """
        self.name = task_name
        self._estimators = None

    def __str__(self) -> str:
        """Name of this task type."""
        return self.name

    @abstractmethod
    def evaluate_model_CV(
        self,
        config: dict,
        estimator: "flaml.automl.ml.BaseEstimator",
        X_train_all: Union[np.ndarray, DataFrame, psDataFrame],
        y_train_all: Union[np.ndarray, DataFrame, Series, psSeries],
        budget: int,
        kf,
        eval_metric: str,
        best_val_loss: float,
        log_training_metric: bool = False,
        fit_kwargs: Optional[dict] = {},
    ) -> Tuple[float, float, float, float]:
        """Evaluate the model using cross-validation.

        Args:
            config: configuration used in the evaluation of the metric.
            estimator: Estimator class of the model.
            X_train_all: Complete training feature data.
            y_train_all: Complete training target data.
            budget: Training time budget.
            kf: Cross-validation index generator.
            eval_metric: Metric name to be used for evaluation.
            best_val_loss: Best current validation-set loss.
            log_training_metric: Bool defaults False. Enables logging of the training metric.
            fit_kwargs: Additional kwargs passed to the estimator's fit method.

        Returns:
            validation loss, metric value, train time, prediction time
        """

    @abstractmethod
    def validate_data(
        self,
        automl: "flaml.automl.automl.AutoML",
        state: "flaml.automl.state.AutoMLState",
        X_train_all: Union[np.ndarray, DataFrame, psDataFrame, None],
        y_train_all: Union[np.ndarray, DataFrame, Series, psSeries, None],
        dataframe: Union[DataFrame, None],
        label: str,
        X_val: Optional[Union[np.ndarray, DataFrame, psDataFrame]] = None,
        y_val: Optional[Union[np.ndarray, DataFrame, Series, psSeries]] = None,
        groups_val: Optional[List[str]] = None,
        groups: Optional[List[str]] = None,
    ):
        """Validate that the data is suitable for this task type.

        Args:
            automl: The AutoML instance from which this task has been constructed.
            state: The AutoMLState instance for this run.
            X_train_all: The complete data set or None if dataframe is supplied.
            y_train_all: The complete target set or None if dataframe is supplied.
            dataframe: A dataframe constaining the complete data set with targets.
            label: The name of the target column in dataframe.
            X_val: Optional. A data set for validation.
            y_val: Optional. A target vector corresponding to X_val for validation.
            groups_val: Group labels (with matching length to y_val) or group counts (with sum equal to length of y_val)
                for validation data. Need to be consistent with groups.
            groups: Group labels (with matching length to y_train) or groups counts (with sum equal to length of y_train)
                for training data.

        Raises:
            AssertionError: The data provided is invalid for this task type and configuration.
        """

    @abstractmethod
    def prepare_data(
        self,
        state: "flaml.automl.state.AutoMLState",
        X_train_all: Union[np.ndarray, DataFrame, psDataFrame],
        y_train_all: Union[np.ndarray, DataFrame, Series, psSeries, None],
        auto_augment: bool,
        eval_method: str,
        split_type: str,
        split_ratio: float,
        n_splits: int,
        data_is_df: bool,
        sample_weight_full: Optional[List[float]] = None,
    ):
        """Prepare the data for fitting or inference.

        Args:
            automl: The AutoML instance from which this task has been constructed.
            state: The AutoMLState instance for this run.
            X_train_all: The complete data set or None if dataframe is supplied. Must
                contain the target if y_train_all is None
            y_train_all: The complete target set or None if supplied in X_train_all.
            auto_augment: If true, task-specific data augmentations will be applied.
            eval_method: A string of resampling strategy, one of ['auto', 'cv', 'holdout'].
            split_type: str or splitter object, default="auto" | the data split type.
                * A valid splitter object is an instance of a derived class of scikit-learn
                [KFold](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.KFold.html#sklearn.model_selection.KFold)
                and have ``split`` and ``get_n_splits`` methods with the same signatures.
                Set eval_method to "cv" to use the splitter object.
                * Valid str options depend on different tasks.
                For classification tasks, valid choices are
                    ["auto", 'stratified', 'uniform', 'time', 'group']. "auto" -> stratified.
                For regression tasks, valid choices are ["auto", 'uniform', 'time', 'group'].
                    "auto" -> uniform.
                For time series forecast tasks, must be "auto" or 'time'.
                For ranking task, must be "auto" or 'group'.
            split_ratio: A float of the valiation data percentage for holdout.
            n_splits: An integer of the number of folds for cross - validation.
            data_is_df: True if the data was provided as a DataFrame else False.
            sample_weight_full: A 1d arraylike of the sample weight.

        Raises:
            AssertionError: The configuration provided is invalid for this task type and data.
        """

    @abstractmethod
    def decide_split_type(
        self,
        split_type: str,
        y_train_all: Union[np.ndarray, DataFrame, Series, psSeries, None],
        fit_kwargs: dict,
        groups: Optional[List[str]] = None,
    ) -> str:
        """Choose an appropriate data split type for this data and task.

        If split_type is 'auto' then this is determined based on the task type and data.
        If a specific split_type is requested then the choice is validated to be appropriate.

        Args:
            split_type: Either 'auto' or a task appropriate split type.
            y_train_all: The complete set of targets.
            fit_kwargs: Additional kwargs passed to the estimator's fit method.
            groups: Optional. Group labels (with matching length to y_train) or groups counts (with sum equal to length
                of y_train) for training data.

        Returns:
            The determined appropriate split type.

        Raises:
            AssertionError: The requested split_type is invalid for this task, configuration and data.
        """

    @abstractmethod
    def preprocess(
        self,
        X: Union[np.ndarray, DataFrame, psDataFrame],
        transformer: Optional["flaml.automl.data.DataTransformer"] = None,
    ) -> Union[np.ndarray, DataFrame]:
        """Preprocess the data ready for fitting or inference with this task type.

        Args:
            X: The data set to process.
            transformer: A DataTransformer instance to be used in processing.

        Returns:
            The preprocessed data set having the same type as the input.
        """

    @abstractmethod
    def default_estimator_list(
        self,
        estimator_list: Union[List[str], str] = "auto",
        is_spark_dataframe: bool = False,
    ) -> List[str]:
        """Return the list of default estimators registered for this task type.

        If 'auto' is provided then the default list is returned, else the provided list will be validated given this task
        type.

        Args:
            estimator_list: Either 'auto' or a list of estimator names to be validated.
            is_spark_dataframe: True if the data is a spark dataframe.

        Returns:
            A list of valid estimator names for this task type.
        """

    @abstractmethod
    def default_metric(self, metric: str) -> str:
        """Return the default metric for this task type.

        If 'auto' is provided then the default metric for this task will be returned. Otherwise, the provided metric name
        is validated for this task type.

        Args:
            metric: The name of a metric to be used in evaluation of models during fitting or validation.

        Returns:
            The default metric, or the provided metric if it is valid for this task type.
        """

    def is_ts_forecast(self) -> bool:
        return self.name in TS_FORECAST

    def is_ts_forecastpanel(self) -> bool:
        return self.name == TS_FORECASTPANEL

    def is_ts_forecastregression(self) -> bool:
        return self.name in TS_FORECASTREGRESSION

    def is_nlp(self) -> bool:
        return self.name in NLP_TASKS

    def is_nlg(self) -> bool:
        return self.name in NLG_TASKS

    def is_classification(self) -> bool:
        return self.name in CLASSIFICATION

    def is_rank(self) -> bool:
        return self.name in RANK

    def is_binary(self) -> bool:
        return self.name == "binary"

    def is_seq_regression(self) -> bool:
        return self.name == SEQREGRESSION

    def is_seq_classification(self) -> bool:
        return self.name == SEQCLASSIFICATION

    def is_token_classification(self) -> bool:
        return self.name == TOKENCLASSIFICATION

    def is_summarization(self) -> bool:
        return self.name == SUMMARIZATION

    def is_multiclass(self) -> bool:
        return "multiclass" in self.name

    def is_regression(self) -> bool:
        return self.name in REGRESSION

    def __eq__(self, other: str) -> bool:
        """For backward compatibility with all the string comparisons to task"""
        return self.name == other

    def estimator_class_from_str(self, estimator_name: str) -> "flaml.automl.ml.BaseEstimator":
        """Determine the estimator class corresponding to the provided name.

        Args:
            estimator_name: Name of the desired estimator.

        Returns:
            The estimator class corresponding to the provided name.

        Raises:
            ValueError: The provided estimator_name has not been registered for this task type.
        """
        if estimator_name in self.estimators:
            return self.estimators[estimator_name]
        else:
            raise ValueError(
                f"{estimator_name} is not a built-in learner for this task type, "
                f"only {list(self.estimators.keys())} are supported."
                "Please use AutoML.add_learner() to add a customized learner."
            )



import argparse
from dataclasses import dataclass, field
from typing import List, Optional

from flaml.automl.task.task import NLG_TASKS

try:
    from transformers import TrainingArguments
except ImportError:
    TrainingArguments = object


@dataclass
class TrainingArgumentsForAuto(TrainingArguments):
    """FLAML custom TrainingArguments.

    Args:
        task (str): the task name for NLP tasks, e.g., seq-classification, token-classification
        output_dir (str): data root directory for outputing the log, etc.
        model_path (str, optional, defaults to "facebook/muppet-roberta-base"): A string,
            the path of the language model file, either a path from huggingface
            model card huggingface.co/models, or a local path for the model.
        fp16 (bool, optional, defaults to "False"): A bool, whether to use FP16.
        max_seq_length (int, optional, defaults to 128): An integer, the max length of the sequence.
            For token classification task, this argument will be ineffective.
        pad_to_max_length (bool, optional, defaults to "False"):
            whether to pad all samples to model maximum sentence length.
            If False, will pad the samples dynamically when batching to the maximum length in the batch.
        per_device_eval_batch_size (int, optional, defaults to 1): An integer, the per gpu evaluation batch size.
        label_list (List[str], optional, defaults to None): A list of string, the string list of the label names.
            When the task is sequence labeling/token classification, there are two formats of the labels:
            (1) The token labels, i.e., [B-PER, I-PER, B-LOC]; (2) Id labels. For (2), need to pass the label_list (e.g., [B-PER, I-PER, B-LOC])
            to convert the Id to token labels when computing the metric with metric_loss_score.
            See the example in [a simple token classification example](/docs/Examples/AutoML-NLP#a-simple-token-classification-example).
    """

    task: str = field(default="seq-classification")

    output_dir: str = field(default="data/output/", metadata={"help": "data dir"})

    model_path: str = field(
        default="facebook/muppet-roberta-base",
        metadata={
            "help": "model path for HPO natural language understanding tasks, default is set to facebook/muppet-roberta-base"
        },
    )

    fp16: bool = field(default=True, metadata={"help": "whether to use the FP16 mode"})

    max_seq_length: int = field(default=128, metadata={"help": "max seq length"})

    label_all_tokens: bool = field(
        default=False,
        metadata={
            "help": "For NER task, whether to set the extra tokenized labels to the same label (instead of -100)"
        },
    )

    pad_to_max_length: bool = field(
        default=False,
        metadata={
            "help": "Whether to pad all samples to model maximum sentence length. "
            "If False, will pad the samples dynamically when batching to the maximum length in the batch. "
        },
    )

    per_device_eval_batch_size: int = field(
        default=1,
        metadata={"help": "per gpu evaluation batch size"},
    )

    label_list: Optional[List[str]] = field(default=None, metadata={"help": "The string list of the label names. "})

    eval_steps: int = field(default=500, metadata={"help": "Run an evaluation every X steps."})

    save_steps: int = field(default=500, metadata={"help": "Save checkpoint every X updates steps."})

    logging_steps: int = field(default=500, metadata={"help": "Log every X updates steps."})

    @staticmethod
    def load_args_from_console():
        from dataclasses import fields

        arg_parser = argparse.ArgumentParser()
        for each_field in fields(TrainingArgumentsForAuto):
            print(each_field)
            arg_parser.add_argument(
                "--" + each_field.name,
                type=each_field.type,
                help=each_field.metadata["help"],
                required=each_field.metadata["required"] if "required" in each_field.metadata else False,
                choices=each_field.metadata["choices"] if "choices" in each_field.metadata else None,
                default=each_field.default,
            )
        console_args, unknown = arg_parser.parse_known_args()
        return console_args


@dataclass
class Seq2SeqTrainingArgumentsForAuto(TrainingArgumentsForAuto):
    model_path: str = field(
        default="t5-small",
        metadata={"help": "model path for HPO natural language generation tasks, default is set to t5-small"},
    )

    sortish_sampler: bool = field(default=False, metadata={"help": "Whether to use SortishSampler or not."})
    predict_with_generate: bool = field(
        default=True,
        metadata={"help": "Whether to use generate to calculate generative metrics (ROUGE, BLEU)."},
    )
    generation_max_length: Optional[int] = field(
        default=None,
        metadata={
            "help": "The `max_length` to use on each evaluation loop when `predict_with_generate=True`. Will default "
            "to the `max_length` value of the model configuration."
        },
    )
    generation_num_beams: Optional[int] = field(
        default=None,
        metadata={
            "help": "The `num_beams` to use on each evaluation loop when `predict_with_generate=True`. Will default "
            "to the `num_beams` value of the model configuration."
        },
    )

    def __post_init__(self):
        super().__post_init__()
        if self.task in NLG_TASKS:
            self.model_path = "t5-small"



import logging
import os
import pathlib
import re
import signal
import subprocess
import sys
import time
from hashlib import md5
from typing import Callable, Dict, List, Optional, Tuple, Union

from flaml.autogen import oai

try:
    import docker
except ImportError:
    docker = None

DEFAULT_MODEL = "gpt-4"
FAST_MODEL = "gpt-3.5-turbo"
# Regular expression for finding a code block
CODE_BLOCK_PATTERN = r"```(\w*)\n(.*?)\n```"
WORKING_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "extensions")
UNKNOWN = "unknown"
TIMEOUT_MSG = "Timeout"
DEFAULT_TIMEOUT = 600


def infer_lang(code):
    """infer the language for the code.
    TODO: make it robust.
    """
    if code.startswith("python ") or code.startswith("pip") or code.startswith("python3 "):
        return "sh"
    return "python"


def extract_code(text: str, pattern: str = CODE_BLOCK_PATTERN) -> List[Tuple[str, str]]:
    """Extract code from a text.

    Args:
        text (str): The text to extract code from.
        pattern (Optional, str): The regular expression pattern for finding the code block.

    Returns:
        list: A list of tuples, each containing the language and the code.
    """
    # Use a regular expression to find all the code blocks
    match = re.findall(pattern, text, flags=re.DOTALL)
    # match = re.search(pattern, text, flags=re.DOTALL)
    # If a match is found, return the code
    # if match:
    #     return match.group(2), match.group(1)
    # If no code block is found, return the whole text
    return match if match else [(UNKNOWN, text)]


# _FIND_CODE_SYS_MSG = [
#     {
#         "role": "system",
#         "content": """In the following conversation, an assistant suggests code and a user is expected to run it.
# Read the conversation, and then find all the right code blocks for the user to run next in the right order.
# Only return the code blocks that are expected to run.
# Don't include code blocks which have been executed unless the user is requested to run the same block again.
# When the user needs to run multiple blocks in sequence, make sure to output all the blocks to run in a right order.
# If the line beginning with "# filename" is put before a code block, move it into the code block as the first line.
# Make sure to add the right "python" or "sh" identifier if the language identifier is missing for a code block.
# Don't make other changes to the code blocks.
# Don't reply anything else if at least one code block is expected to run.
# If no code block is expeted to run, check whether the task has been successfully finished at full satisfaction.
# If not, reply with the reason why the task is not finished.""",
#     },
# ]
# _FIND_CODE_CONFIG = {
#     "model": FAST_MODEL,
# }


# def find_code(messages: List[Dict], sys_msg=None, **config) -> Tuple[List[Tuple[str, str]], str]:
#     """Find code from a list of messages.

#     Args:
#         messages (str): The list of messages to find code from.
#         sys_msg (Optional, str): The system message to prepend to the messages.
#         config (Optional, dict): The configuration for the API call.

#     Returns:
#         list: A list of tuples, each containing the language and the code.
#         str: The generated text by llm.
#     """
#     params = {**_FIND_CODE_CONFIG, **config}
#     if sys_msg is None or not sys_msg[0]["content"]:
#         sys_msg = _FIND_CODE_SYS_MSG
#     response = oai.ChatCompletion.create(messages=sys_msg + messages, **params)
#     content = oai.Completion.extract_text(response)[0]
#     return extract_code(content), content


def generate_code(pattern: str = CODE_BLOCK_PATTERN, **config) -> Tuple[str, float]:
    """Generate code.

    Args:
        pattern (Optional, str): The regular expression pattern for finding the code block.
            The default pattern is for finding a code block in a markdown file.
        config (Optional, dict): The configuration for the API call.

    Returns:
        str: The generated code.
        float: The cost of the generation.
    """
    response = oai.Completion.create(**config)
    return extract_code(oai.Completion.extract_text(response)[0], pattern), response["cost"]


_IMPROVE_FUNCTION_CONFIG = {
    "prompt": """Improve the function '{func_name}' to achieve the objective '{objective}'.
The current implementation of the function is as follows:
{file_string}""",
    "model": DEFAULT_MODEL,
    "request_timeout": 600,
}


def improve_function(file_name, func_name, objective, **config):
    """(work in progress) Improve the function to achieve the objective."""
    params = {**_IMPROVE_FUNCTION_CONFIG, **config}
    # read the entire file into a str
    with open(file_name) as f:
        file_string = f.read()
    response = oai.Completion.create(
        {"func_name": func_name, "objective": objective, "file_string": file_string}, **params
    )
    return oai.Completion.extract_text(response)[0], response["cost"]


_IMPROVE_CODE_CONFIG = {
    "prompt": """Analyze the code in the following files and return a list of suggestions for improvement{followup}, to achieve the objective of '{objective}'.
{code}
""",
    "model": DEFAULT_MODEL,
    "request_timeout": 900,
}


def improve_code(files, objective, suggest_only=True, **config):
    """Improve the code to achieve a given objective.

    Args:
        files (list): A list of file names containing the source code.
        objective (str): The objective to achieve.
        suggest_only (bool): Whether to return only the suggestions or the improved code.
        config (Optional, dict): The configuration for the API call.

    Returns:
        str: The improved code if suggest_only=False; a list of suggestions if suggest_only=True (default).
        float: The cost of the generation.
    """
    code = ""
    for file_name in files:
        # read the entire file into a string
        with open(file_name) as f:
            file_string = f.read()
        code += f"""{file_name}:
{file_string}

"""
    params = {**_IMPROVE_CODE_CONFIG, **config}
    followup = "" if suggest_only else " followed by the improved code"
    response = oai.Completion.create({"objective": objective, "code": code, "followup": followup}, **params)
    return oai.Completion.extract_text(response)[0], response["cost"]


def timeout_handler(signum, frame):
    raise TimeoutError("Timed out!")


def _cmd(lang):
    if lang.startswith("python") or lang in ["bash", "sh"]:
        return lang
    if lang == "shell":
        return "sh"
    raise NotImplementedError(f"{lang} not recognized in code execution")


def execute_code(
    code: Optional[str] = None,
    timeout: Optional[int] = None,
    filename: Optional[str] = None,
    work_dir: Optional[str] = None,
    use_docker: Optional[Union[List[str], str, bool]] = docker is not None,
    lang: Optional[str] = "python",
) -> Tuple[int, str, str]:
    """Execute code in a docker container.
    This function is not tested on MacOS.

    Args:
        code (Optional, str): The code to execute.
            If None, the code from the file specified by filename will be executed.
            Either code or filename must be provided.
        timeout (Optional, int): The maximum execution time in seconds.
            If None, a default timeout will be used. The default timeout is 600 seconds. On Windows, the timeout is not enforced when use_docker=False.
        filename (Optional, str): The file name to save the code or where the code is stored when `code` is None.
            If None, a file with a randomly generated name will be created.
            The randomly generated file will be deleted after execution.
            The file name must be a relative path. Relative paths are relative to the working directory.
        work_dir (Optional, str): The working directory for the code execution.
            If None, a default working directory will be used.
            The default working directory is the "extensions" directory under
            "path_to_flaml/autogen".
        use_docker (Optional, list, str or bool): The docker image to use for code execution.
            If a list or a str of image name(s) is provided, the code will be executed in a docker container
            with the first image successfully pulled.
            If None, False or empty, the code will be executed in the current environment.
            Default is True, which will be converted into a list.
            If the code is executed in the current environment,
            the code must be trusted.
        lang (Optional, str): The language of the code. Default is "python".

    Returns:
        int: 0 if the code executes successfully.
        str: The error message if the code fails to execute; the stdout otherwise.
        image: The docker image name after container run when docker is used.
    """
    assert code is not None or filename is not None, "Either code or filename must be provided."
    timeout = timeout or DEFAULT_TIMEOUT
    original_filename = filename
    if filename is None:
        code_hash = md5(code.encode()).hexdigest()
        # create a file with a automatically generated name
        filename = f"tmp_code_{code_hash}.{'py' if lang.startswith('python') else lang}"
    if work_dir is None:
        work_dir = WORKING_DIR
    filepath = os.path.join(work_dir, filename)
    file_dir = os.path.dirname(filepath)
    os.makedirs(file_dir, exist_ok=True)
    if code is not None:
        with open(filepath, "w") as fout:
            fout.write(code)
    # check if already running in a docker container
    in_docker_container = os.path.exists("/.dockerenv")
    if not use_docker or in_docker_container:
        # already running in a docker container
        cmd = [sys.executable if lang.startswith("python") else _cmd(lang), filename]
        if sys.platform == "win32":
            logging.warning("SIGALRM is not supported on Windows. No timeout will be enforced.")
            result = subprocess.run(
                cmd,
                cwd=work_dir,
                capture_output=True,
            )
        else:
            signal.signal(signal.SIGALRM, timeout_handler)
            try:
                signal.alarm(timeout)
                # run the code in a subprocess in the current docker container in the working directory
                result = subprocess.run(
                    cmd,
                    cwd=work_dir,
                    capture_output=True,
                )
                signal.alarm(0)
            except TimeoutError:
                if original_filename is None:
                    os.remove(filepath)
                return 1, TIMEOUT_MSG, None
        if original_filename is None:
            os.remove(filepath)
            abs_path = str(pathlib.Path(filepath).absolute())
        else:
            abs_path = str(pathlib.Path(work_dir).absolute()) + "/"
        if result.returncode:
            logs = result.stderr.decode("utf-8")
            logs = logs.replace(str(abs_path), "")
        else:
            logs = result.stdout.decode("utf-8")
        return result.returncode, logs, None

    # create a docker client
    client = docker.from_env()
    image_list = (
        ["python:3-alpine", "python:3", "python:3-windowsservercore"]
        if use_docker is True
        else [use_docker]
        if isinstance(use_docker, str)
        else use_docker
    )
    for image in image_list:
        # check if the image exists
        try:
            client.images.get(image)
            break
        except docker.errors.ImageNotFound:
            # pull the image
            print("Pulling image", image)
            try:
                client.images.pull(image)
                break
            except docker.errors.DockerException:
                print("Failed to pull image", image)
    # get a randomized str based on current time to wrap the exit code
    exit_code_str = f"exitcode{time.time()}"
    abs_path = pathlib.Path(work_dir).absolute()
    # if sys.platform == "win32":
    #     abs_path = str(abs_path).replace("\\", "/")
    #     abs_path = f"/{abs_path[0].lower()}{abs_path[2:]}"
    cmd = [
        "sh",
        "-c",
        f"{_cmd(lang)} {filename}; exit_code=$?; echo -n {exit_code_str}; echo -n $exit_code; echo {exit_code_str}",
    ]
    # create a docker container
    container = client.containers.run(
        image,
        command=cmd,
        working_dir="/workspace",
        detach=True,
        # get absolute path to the working directory
        volumes={abs_path: {"bind": "/workspace", "mode": "rw"}},
    )
    start_time = time.time()
    while container.status != "exited" and time.time() - start_time < timeout:
        # Reload the container object
        container.reload()
    if container.status != "exited":
        container.stop()
        container.remove()
        if original_filename is None:
            os.remove(filepath)
        return 1, TIMEOUT_MSG, image
    # try:
    #     container.wait(timeout=timeout)
    # except (ReadTimeout, ConnectionError):
    #     container.stop()
    #     container.remove()
    #     if original_filename is None:
    #         os.remove(filepath)
    #     return 1, "Timeout"
    # get the container logs
    logs = container.logs().decode("utf-8").rstrip()
    # commit the image
    tag = filename.replace("/", "")
    container.commit(repository="python", tag=tag)
    # remove the container
    container.remove()
    # check if the code executed successfully
    exit_code = container.attrs["State"]["ExitCode"]
    if exit_code == 0:
        # extract the exit code from the logs
        pattern = re.compile(f"{exit_code_str}(\\d+){exit_code_str}")
        match = pattern.search(logs)
        exit_code = 1 if match is None else int(match.group(1))
        # remove the exit code from the logs
        logs = logs if match is None else pattern.sub("", logs)

    if original_filename is None:
        os.remove(filepath)
    if exit_code:
        logs = logs.replace(f"/workspace/{filename if original_filename is None else ''}", "")
    # return the exit code, logs and image
    return exit_code, logs, f"python:{tag}"


_GENERATE_ASSERTIONS_CONFIG = {
    "prompt": """Given the signature and docstring, write the exactly same number of assertion(s) for the provided example(s) in the docstring, without assertion messages.

func signature:
{definition}
assertions:""",
    "model": FAST_MODEL,
    "max_tokens": 256,
    "stop": "\n\n",
}


def generate_assertions(definition: str, **config) -> Tuple[str, float]:
    """Generate assertions for a function.

    Args:
        definition (str): The function definition, including the signature and docstr.
        config (Optional, dict): The configuration for the API call.

    Returns:
        str: The generated assertions.
        float: The cost of the generation.
    """
    params = {**_GENERATE_ASSERTIONS_CONFIG, **config}
    response = oai.Completion.create(
        {"definition": definition},
        **params,
    )
    assertions = oai.Completion.extract_text(response)[0]
    return assertions, response["cost"]


def _remove_check(response):
    """Remove the check function from the response."""
    # find the position of the check function
    pos = response.find("def check(")
    if pos == -1:
        return response
    return response[:pos]


def eval_function_completions(
    responses: List[str],
    definition: str,
    test: Optional[str] = None,
    entry_point: Optional[str] = None,
    assertions: Optional[Union[str, Callable[[str], Tuple[str, float]]]] = None,
    timeout: Optional[float] = 3,
    use_docker: Optional[bool] = True,
) -> Dict:
    """Select a response from a list of responses for the function completion task (using generated assertions), and/or evaluate if the task is successful using a gold test.

    Args:
        responses (list): The list of responses.
        definition (str): The input definition.
        test (Optional, str): The test code.
        entry_point (Optional, str): The name of the function.
        assertions (Optional, str or Callable): The assertion code which serves as a filter of the responses, or an assertion generator.
            When provided, only the responses that pass the assertions will be considered for the actual test (if provided).
        timeout (Optional, float): The timeout for executing the code.

    Returns:
        dict: The success metrics.
    """
    n = len(responses)
    if assertions is None:
        # no assertion filter
        success_list = []
        for i in range(n):
            response = _remove_check(responses[i])
            code = (
                f"{response}\n{test}\ncheck({entry_point})"
                if response.startswith("def")
                else f"{definition}{response}\n{test}\ncheck({entry_point})"
            )
            success = execute_code(code, timeout=timeout, use_docker=use_docker)[0] == 0
            success_list.append(success)
        return {
            "expected_success": 1 - pow(1 - sum(success_list) / n, n),
            "success": any(s for s in success_list),
        }
    if callable(assertions) and n > 1:
        # assertion generator
        assertions, gen_cost = assertions(definition)
    else:
        gen_cost = 0
    if n > 1 or test is None:
        for i in range(n):
            response = responses[i] = _remove_check(responses[i])
            code = (
                f"{response}\n{assertions}" if response.startswith("def") else f"{definition}{response}\n{assertions}"
            )
            succeed_assertions = execute_code(code, timeout=timeout, use_docker=use_docker)[0] == 0
            if succeed_assertions:
                break
    else:
        # just test, no need to check assertions
        succeed_assertions = False
        i, response = 0, responses[0]
    if test is None:
        # no test code
        return {
            "index_selected": i,
            "succeed_assertions": succeed_assertions,
            "gen_cost": gen_cost,
            "assertions": assertions,
        }
    code_test = (
        f"{response}\n{test}\ncheck({entry_point})"
        if response.startswith("def")
        else f"{definition}{response}\n{test}\ncheck({entry_point})"
    )
    success = execute_code(code_test, timeout=timeout, use_docker=use_docker)[0] == 0
    return {
        "index_selected": i,
        "succeed_assertions": succeed_assertions,
        "success": success,
        "gen_cost": gen_cost,
        "assertions": assertions,
    }


_FUNC_COMPLETION_PROMPT = "# Python 3{definition}"
_FUNC_COMPLETION_STOP = ["\nclass", "\ndef", "\nif", "\nprint"]
_IMPLEMENT_CONFIGS = [
    {"model": FAST_MODEL, "prompt": _FUNC_COMPLETION_PROMPT, "temperature": 0, "seed": 0},
    {"model": FAST_MODEL, "prompt": _FUNC_COMPLETION_PROMPT, "stop": _FUNC_COMPLETION_STOP, "n": 7, "seed": 0},
    {"model": DEFAULT_MODEL, "prompt": _FUNC_COMPLETION_PROMPT, "temperature": 0, "seed": 1},
    {"model": DEFAULT_MODEL, "prompt": _FUNC_COMPLETION_PROMPT, "stop": _FUNC_COMPLETION_STOP, "n": 2, "seed": 2},
    {"model": DEFAULT_MODEL, "prompt": _FUNC_COMPLETION_PROMPT, "stop": _FUNC_COMPLETION_STOP, "n": 1, "seed": 2},
]


class PassAssertionFilter:
    def __init__(self, assertions):
        self._assertions = assertions
        self.cost = 0
        self.metrics = self.responses = None

    def pass_assertions(self, context, response, **_):
        """Check if the response passes the assertions."""
        responses = oai.Completion.extract_text(response)
        metrics = eval_function_completions(responses, context["definition"], assertions=self._assertions)
        self._assertions = metrics["assertions"]
        self.cost += metrics["gen_cost"]
        self.metrics = metrics
        self.responses = responses
        return metrics["succeed_assertions"]


def implement(
    definition: str,
    configs: Optional[List[Dict]] = None,
    assertions: Optional[Union[str, Callable[[str], Tuple[str, float]]]] = generate_assertions,
) -> Tuple[str, float]:
    """Implement a function from a definition.

    Args:
        definition (str): The function definition, including the signature and docstr.
        configs (list): The list of configurations for completion.
        assertions (Optional, str or Callable): The assertion code which serves as a filter of the responses, or an assertion generator.

    Returns:
        str: The implementation.
        float: The cost of the implementation.
        int: The index of the configuration which generates the implementation.
    """
    cost = 0
    configs = configs or _IMPLEMENT_CONFIGS
    if len(configs) > 1 and callable(assertions):
        assertions, cost = assertions(definition)
    assertion_filter = PassAssertionFilter(assertions)
    response = oai.Completion.create(
        {"definition": definition}, config_list=configs, filter_func=assertion_filter.pass_assertions
    )
    cost += assertion_filter.cost + response["cost"]
    return assertion_filter.responses[assertion_filter.metrics["index_selected"]], cost, response["config_id"]

    # for i, config in enumerate(configs):
    #     response = oai.Completion.create({"definition": definition}, **config)
    #     cost += oai.Completion.cost(response)
    #     responses = oai.Completion.extract_text(response)
    #     metrics = eval_function_completions(responses, definition, assertions=assertions)
    #     assertions = metrics["assertions"]
    #     cost += metrics["gen_cost"]
    #     if metrics["succeed_assertions"] or i == len(configs) - 1:
    #         return responses[metrics["index_selected"]], cost, i



import logging
from typing import List, Optional, Tuple, Union

import numpy as np

from flaml.automl.spark import (
    DataFrame,
    F,
    Series,
    T,
    _spark_major_minor_version,
    ps,
    psDataFrame,
    psSeries,
    set_option,
    sparkDataFrame,
)

logger = logging.getLogger(__name__)
logger_formatter = logging.Formatter(
    "[%(name)s: %(asctime)s] {%(lineno)d} %(levelname)s - %(message)s", "%m-%d %H:%M:%S"
)
logger.propagate = False


def to_pandas_on_spark(
    df: Union[DataFrame, sparkDataFrame, Series, psDataFrame, psSeries],
    index_col: Optional[str] = None,
    default_index_type: Optional[str] = "distributed-sequence",
) -> Union[psDataFrame, psSeries]:
    """Convert pandas or pyspark dataframe/series to pandas_on_Spark dataframe/series.

    Args:
        df: pandas.DataFrame/series or pyspark dataframe | The input dataframe/series.
        index_col: str, optional | The column name to use as index, default None.
        default_index_type: str, optional | The default index type, default "distributed-sequence".

    Returns:
        pyspark.pandas.DataFrame/Series: The converted pandas-on-Spark dataframe/series.

    ```python
    import pandas as pd
    from flaml.automl.spark.utils import to_pandas_on_spark

    pdf = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    psdf = to_pandas_on_spark(pdf)
    print(psdf)

    from pyspark.sql import SparkSession

    spark = SparkSession.builder.getOrCreate()
    sdf = spark.createDataFrame(pdf)
    psdf = to_pandas_on_spark(sdf)
    print(psdf)

    pds = Series([1, 2, 3])
    pss = to_pandas_on_spark(pds)
    print(pss)
    ```
    """
    set_option("compute.default_index_type", default_index_type)
    if isinstance(df, (DataFrame, Series)):
        return ps.from_pandas(df)
    elif isinstance(df, sparkDataFrame):
        if _spark_major_minor_version[0] == 3 and _spark_major_minor_version[1] < 3:
            return df.to_pandas_on_spark(index_col=index_col)
        else:
            return df.pandas_api(index_col=index_col)
    elif isinstance(df, (psDataFrame, psSeries)):
        return df
    else:
        raise TypeError(f"{type(df)} is not one of pandas.DataFrame, pandas.Series and pyspark.sql.DataFrame")


def train_test_split_pyspark(
    df: Union[sparkDataFrame, psDataFrame],
    stratify_column: Optional[str] = None,
    test_fraction: Optional[float] = 0.2,
    seed: Optional[int] = 1234,
    to_pandas_spark: Optional[bool] = True,
    index_col: Optional[str] = "tmp_index_col",
) -> Tuple[Union[sparkDataFrame, psDataFrame], Union[sparkDataFrame, psDataFrame]]:
    """Split a pyspark dataframe into train and test dataframes.

    Args:
        df: pyspark.sql.DataFrame | The input dataframe.
        stratify_column: str | The column name to stratify the split. Default None.
        test_fraction: float | The fraction of the test data. Default 0.2.
        seed: int | The random seed. Default 1234.
        to_pandas_spark: bool | Whether to convert the output to pandas_on_spark. Default True.
        index_col: str | The column name to use as index. Default None.

    Returns:
        pyspark.sql.DataFrame/pandas_on_spark DataFrame | The train dataframe.
        pyspark.sql.DataFrame/pandas_on_spark DataFrame | The test dataframe.
    """
    import warnings

    warnings.filterwarnings("ignore")

    if isinstance(df, psDataFrame):
        df = df.to_spark(index_col=index_col)

    if stratify_column:
        # Test data
        test_fraction_dict = (
            df.select(stratify_column).distinct().withColumn("fraction", F.lit(test_fraction)).rdd.collectAsMap()
        )
        df_test = df.stat.sampleBy(stratify_column, test_fraction_dict, seed)
        # Train data
        df_train = df.subtract(df_test)
    else:
        df_train, df_test = df.randomSplit([1 - test_fraction, test_fraction], seed)

    if to_pandas_spark:
        df_train = to_pandas_on_spark(df_train, index_col=index_col)
        df_test = to_pandas_on_spark(df_test, index_col=index_col)
        df_train.index.name = None
        df_test.index.name = None
    elif index_col == "tmp_index_col":
        df_train = df_train.drop(index_col)
        df_test = df_test.drop(index_col)
    return [df_train, df_test]


def unique_pandas_on_spark(psds: Union[psSeries, psDataFrame]) -> Tuple[np.ndarray, np.ndarray]:
    """Get the unique values and counts of a pandas_on_spark series."""
    if isinstance(psds, psDataFrame):
        psds = psds.iloc[:, 0]
    _tmp = psds.value_counts().to_pandas()
    label_set = _tmp.index.values
    counts = _tmp.values
    return label_set, counts


def len_labels(y: Union[psSeries, np.ndarray], return_labels=False) -> Union[int, Optional[np.ndarray]]:
    """Get the number of unique labels in y."""
    if not isinstance(y, (psDataFrame, psSeries)):
        labels = np.unique(y)
    else:
        labels = y.unique() if isinstance(y, psSeries) else y.iloc[:, 0].unique()
    if return_labels:
        return len(labels), labels
    return len(labels)


def unique_value_first_index(y: Union[Series, psSeries, np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
    """Get the unique values and indices of a pandas series,
    pandas_on_spark series or numpy array."""
    if isinstance(y, psSeries):
        y_unique = y.drop_duplicates().sort_index()
        label_set = y_unique.values
        first_index = y_unique.index.values
    else:
        label_set, first_index = np.unique(y, return_index=True)
    return label_set, first_index


def iloc_pandas_on_spark(
    psdf: Union[psDataFrame, psSeries, DataFrame, Series],
    index: Union[int, slice, list],
    index_col: Optional[str] = "tmp_index_col",
) -> Union[psDataFrame, psSeries]:
    """Get the rows of a pandas_on_spark dataframe/series by index."""
    import warnings

    warnings.filterwarnings("ignore")

    if isinstance(psdf, (DataFrame, Series)):
        return psdf.iloc[index]
    if isinstance(index, (int, slice)):
        if isinstance(psdf, psSeries):
            return psdf.iloc[index]
        else:
            return psdf.iloc[index, :]
    elif isinstance(index, list):
        if isinstance(psdf, psSeries):
            sdf = psdf.to_frame().to_spark(index_col=index_col)
        else:
            if index_col not in psdf.columns:
                sdf = psdf.to_spark(index_col=index_col)
            else:
                sdf = psdf.to_spark()
        sdfiloc = sdf.filter(F.col(index_col).isin(index))
        psdfiloc = to_pandas_on_spark(sdfiloc)
        if isinstance(psdf, psSeries):
            psdfiloc = psdfiloc[psdfiloc.columns.drop(index_col)[0]]
        elif index_col not in psdf.columns:
            psdfiloc = psdfiloc.drop(columns=[index_col])
        return psdfiloc
    else:
        raise TypeError(f"{type(index)} is not one of int, slice and list for pandas_on_spark iloc")


def spark_kFold(
    dataset: Union[sparkDataFrame, psDataFrame],
    nFolds: int = 3,
    foldCol: str = "",
    seed: int = 42,
    index_col: Optional[str] = "tmp_index_col",
) -> List[Tuple[psDataFrame, psDataFrame]]:
    """Generate k-fold splits for a Spark DataFrame.
    Adopted from https://spark.apache.org/docs/latest/api/python/_modules/pyspark/ml/tuning.html#CrossValidator

    Args:
        dataset: sparkDataFrame / psDataFrame. | The DataFrame to split.
        nFolds: int | The number of folds. Default is 3.
        foldCol: str | The column name to use for fold numbers. If not specified,
            the DataFrame will be randomly split. Default is "".
            The same group will not appear in two different folds (the number of
            distinct groups has to be at least equal to the number of folds).
            The folds are approximately balanced in the sense that the number of
            distinct groups is approximately the same in each fold.
        seed: int | The random seed. Default is 42.
        index_col: str | The name of the index column. Default is "tmp_index_col".

    Returns:
        A list of (train, validation) DataFrames.
    """
    import warnings

    warnings.filterwarnings("ignore")

    if isinstance(dataset, psDataFrame):
        dataset = dataset.to_spark(index_col=index_col)

    datasets = []
    if not foldCol:
        # Do random k-fold split.
        h = 1.0 / nFolds
        randCol = f"rand_col_{seed}"
        df = dataset.select("*", F.rand(seed).alias(randCol))
        for i in range(nFolds):
            validateLB = i * h
            validateUB = (i + 1) * h
            condition = (df[randCol] >= validateLB) & (df[randCol] < validateUB)
            validation = to_pandas_on_spark(df.filter(condition), index_col=index_col)
            train = to_pandas_on_spark(df.filter(~condition), index_col=index_col)
            datasets.append((train.drop(columns=[randCol]), validation.drop(columns=[randCol])))
    else:
        # Use user-specified fold column
        def get_fold_num(foldNum: int) -> int:
            return int(foldNum % nFolds)

        get_fold_num_udf = F.UserDefinedFunction(get_fold_num, T.IntegerType())
        for i in range(nFolds):
            training = dataset.filter(get_fold_num_udf(dataset[foldCol]) != F.lit(i))
            validation = dataset.filter(get_fold_num_udf(dataset[foldCol]) == F.lit(i))
            if training.rdd.getNumPartitions() == 0 or len(training.take(1)) == 0:
                raise ValueError("The training data at fold %s is empty." % i)
            if validation.rdd.getNumPartitions() == 0 or len(validation.take(1)) == 0:
                raise ValueError("The validation data at fold %s is empty." % i)
            training = to_pandas_on_spark(training, index_col=index_col)
            validation = to_pandas_on_spark(validation, index_col=index_col)
            datasets.append((training, validation))

    return datasets



import logging
import os
import textwrap
import threading
import time
from functools import lru_cache, partial

logger = logging.getLogger(__name__)
logger_formatter = logging.Formatter(
    "[%(name)s: %(asctime)s] {%(lineno)d} %(levelname)s - %(message)s", "%m-%d %H:%M:%S"
)
logger.propagate = False
os.environ["PYARROW_IGNORE_TIMEZONE"] = "1"
try:
    import py4j
    import pyspark
    from pyspark.sql import SparkSession
    from pyspark.util import VersionUtils
except ImportError:
    _have_spark = False
    py4j = None
    _spark_major_minor_version = (0, 0)
else:
    _have_spark = True
    _spark_major_minor_version = VersionUtils.majorMinorVersion(pyspark.__version__)


@lru_cache(maxsize=2)
def check_spark():
    """Check if Spark is installed and running.
    Result of the function will be cached since test once is enough. As lru_cache will not
    cache exceptions, we don't raise exceptions here but only log a warning message.

    Returns:
        Return (True, None) if the check passes, otherwise log the exception message and
        return (False, Exception(msg)). The exception can be raised by the caller.
    """
    logger.debug("\nchecking Spark installation...This line should appear only once.\n")
    if not _have_spark:
        msg = """use_spark=True requires installation of PySpark. Please run pip install flaml[spark]
        and check [here](https://spark.apache.org/docs/latest/api/python/getting_started/install.html)
        for more details about installing Spark."""
        return False, ImportError(msg)

    if _spark_major_minor_version[0] < 3:
        msg = "Spark version must be >= 3.0 to use flaml[spark]"
        return False, ImportError(msg)

    try:
        SparkSession.builder.getOrCreate()
    except RuntimeError as e:
        return False, RuntimeError(e)

    return True, None


def get_n_cpus(node="driver"):
    """Get the number of CPU cores of the given type of node.

    Args:
        node: string | The type of node to get the number of cores. Can be 'driver' or 'executor'.
            Default is 'driver'.

    Returns:
        An int of the number of CPU cores.
    """
    assert node in ["driver", "executor"]
    try:
        n_cpus = int(SparkSession.builder.getOrCreate().sparkContext.getConf().get(f"spark.{node}.cores"))
    except (TypeError, RuntimeError):
        n_cpus = os.cpu_count()
    return n_cpus


def with_parameters(trainable, **kwargs):
    """Wrapper for trainables to pass arbitrary large data objects.

    This wrapper function will store all passed parameters in the Spark
    Broadcast variable.

    Args:
        trainable: Trainable to wrap.
        **kwargs: parameters to store in object store.

    Returns:
        A new function with partial application of the given arguments
        and keywords. The given arguments and keywords will be broadcasted
        to all the executors.


    ```python
    import pyspark
    import flaml
    from sklearn.datasets import load_iris
    def train(config, data=None):
        if isinstance(data, pyspark.broadcast.Broadcast):
            data = data.value
        print(config, data)

    data = load_iris()
    with_parameters_train = flaml.tune.spark.utils.with_parameters(train, data=data)
    with_parameters_train(config=1)
    train(config={"metric": "accuracy"})
    ```
    """

    if not callable(trainable):
        raise ValueError(
            f"`with_parameters() only works with function trainables`. " f"Got type: " f"{type(trainable)}."
        )

    spark_available, spark_error_msg = check_spark()
    if not spark_available:
        raise spark_error_msg
    spark = SparkSession.builder.getOrCreate()

    bc_kwargs = dict()
    for k, v in kwargs.items():
        bc_kwargs[k] = spark.sparkContext.broadcast(v)

    return partial(trainable, **bc_kwargs)


def broadcast_code(custom_code="", file_name="mylearner"):
    """Write customized learner/metric code contents to a file for importing.
    It is necessary for using the customized learner/metric in spark backend.
    The path of the learner/metric file will be returned.

    Args:
        custom_code: str, default="" | code contents of the custom learner/metric.
        file_name: str, default="mylearner" | file name of the custom learner/metric.

    Returns:
        The path of the custom code file.
    ```python
    from flaml.tune.spark.utils import broadcast_code
    from flaml.automl.model import LGBMEstimator

    custom_code = '''
    from flaml.automl.model import LGBMEstimator
    from flaml import tune

    class MyLargeLGBM(LGBMEstimator):
        @classmethod
        def search_space(cls, **params):
            return {
                "n_estimators": {
                    "domain": tune.lograndint(lower=4, upper=32768),
                    "init_value": 32768,
                    "low_cost_init_value": 4,
                },
                "num_leaves": {
                    "domain": tune.lograndint(lower=4, upper=32768),
                    "init_value": 32768,
                    "low_cost_init_value": 4,
                },
            }
    '''

    broadcast_code(custom_code=custom_code)
    from flaml.tune.spark.mylearner import MyLargeLGBM
    assert isinstance(MyLargeLGBM(), LGBMEstimator)
    ```
    """
    # Check if Spark is available
    spark_available, _ = check_spark()

    # Write to local driver file system
    flaml_path = os.path.dirname(os.path.abspath(__file__))
    custom_code = textwrap.dedent(custom_code)
    custom_path = os.path.join(flaml_path, file_name + ".py")

    with open(custom_path, "w") as f:
        f.write(custom_code)

    # If using Spark, broadcast the code content to executors
    if spark_available:
        spark = SparkSession.builder.getOrCreate()
        bc_code = spark.sparkContext.broadcast(custom_code)

        # Execute a job to ensure the code is distributed to all executors
        def _write_code(bc):
            code = bc.value
            import os

            module_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), file_name + ".py")
            os.makedirs(os.path.dirname(module_path), exist_ok=True)
            with open(module_path, "w") as f:
                f.write(code)
            return True

        spark.sparkContext.parallelize(range(1)).map(lambda _: _write_code(bc_code)).collect()

    return custom_path


def get_broadcast_data(broadcast_data):
    """Get the broadcast data from the broadcast variable.

    Args:
        broadcast_data: pyspark.broadcast.Broadcast | the broadcast variable.

    Returns:
        The broadcast data.
    """
    if _have_spark and isinstance(broadcast_data, pyspark.broadcast.Broadcast):
        broadcast_data = broadcast_data.value
    return broadcast_data


class PySparkOvertimeMonitor:
    """A context manager class to monitor if the PySpark job is overtime.
    Example:

    ```python
    with PySparkOvertimeMonitor(time_start, time_budget_s, force_cancel, parallel=parallel):
        results = parallel(
            delayed(evaluation_function)(trial_to_run.config)
            for trial_to_run in trials_to_run
        )
    ```

    """

    def __init__(
        self,
        start_time,
        time_budget_s,
        force_cancel=False,
        cancel_func=None,
        parallel=None,
        sc=None,
    ):
        """Constructor.

        Specify the time budget and start time of the PySpark job, and specify how to cancel them.

        Args:
            Args relate to monitoring:
                start_time: float | The start time of the PySpark job.
                time_budget_s: float | The time budget of the PySpark job in seconds.
                force_cancel: boolean, default=False | Whether to forcely cancel the PySpark job if overtime.

            Args relate to how to cancel the PySpark job:
            (Only one of the following args will work. Priorities from top to bottom)
                cancel_func: function | A function to cancel the PySpark job.
                parallel: joblib.parallel.Parallel | Specify this if using joblib_spark as a parallel backend. It will call parallel._backend.terminate() to cancel the jobs.
                sc: pyspark.SparkContext object | You can pass a specific SparkContext.

                If all three args is None, the monitor will call pyspark.SparkContext.getOrCreate().cancelAllJobs() to cancel the jobs.


        """
        self._time_budget_s = time_budget_s
        self._start_time = start_time
        self._force_cancel = force_cancel
        # TODO: add support for non-spark scenario
        if self._force_cancel and _have_spark:
            self._monitor_daemon = None
            self._finished_flag = False
            self._cancel_flag = False
            self.sc = None
            if cancel_func:
                self.__cancel_func = cancel_func
            elif parallel:
                self.__cancel_func = parallel._backend.terminate
            elif sc:
                self.sc = sc
                self.__cancel_func = self.sc.cancelAllJobs
            else:
                self.__cancel_func = pyspark.SparkContext.getOrCreate().cancelAllJobs
            # logger.info(self.__cancel_func)

    def _monitor_overtime(self):
        """The lifecycle function for monitor thread."""
        if self._time_budget_s is None:
            self.__cancel_func()
            self._cancel_flag = True
            return
        while time.time() - self._start_time <= self._time_budget_s:
            time.sleep(0.01)
            if self._finished_flag:
                return
        self.__cancel_func()
        self._cancel_flag = True
        return

    def _setLogLevel(self, level):
        """Set the log level of the spark context.
        Set the level to OFF could block the warning message of Spark."""
        if self.sc:
            self.sc.setLogLevel(level)
        else:
            pyspark.SparkContext.getOrCreate().setLogLevel(level)

    def __enter__(self):
        """Enter the context manager.
        This will start a monitor thread if spark is available and force_cancel is True.
        """
        if self._force_cancel and _have_spark:
            self._monitor_daemon = threading.Thread(target=self._monitor_overtime)
            # logger.setLevel("INFO")
            logger.info("monitor started")
            self._setLogLevel("OFF")
            self._monitor_daemon.start()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Exit the context manager.
        This will wait for the monitor thread to nicely exit."""
        logger.debug(f"monitor exited: {exc_type}, {exc_value}, {exc_traceback}")
        if self._force_cancel and _have_spark:
            self._finished_flag = True
            self._monitor_daemon.join()
            if self._cancel_flag:
                print()
                logger.warning("Time exceeded, canceled jobs")
            # self._setLogLevel("WARN")
            if not exc_type:
                return True
            elif exc_type == py4j.protocol.Py4JJavaError:
                logger.debug("Py4JJavaError Exception: %s", exc_value)
                return True
            elif exc_type == TypeError:
                # When force cancel, joblib>1.2.0 will raise joblib.externals.loky.process_executor._ExceptionWithTraceback
                logger.debug("TypeError Exception: %s", exc_value)
                return True
            else:
                return False
